"""

#######################################################################
A simple agent framework which you can extend with some custom logic.

As-is, this agent will faucet itself some tokens in the setup phase
and then do nothing for the rest of the trading session.

Fill in your own logic into the `step` function to make them trade
however you'd like.

Below, we have a range of building blocks,
copy and paste these into your code to get started

#######################################################################

### Pull best bid/ask prices
best_bid, best_ask = self.vega.best_prices(self.market_id)

### Pull market depth (up to a specified number of levels)
market_depth = self.vega.market_depth(self.market_id, num_levels=5)

### Place a Limit Order
self.vega.submit_order(
    trading_wallet=self.wallet_name,
    market_id=self.market_id,
    order_type="TYPE_LIMIT",
    # https://docs.vega.xyz/docs/mainnet/grpc/vega/vega.proto#ordertimeinforce
    time_in_force="TIME_IN_FORCE_GTC",
    # https://docs.vega.xyz/docs/mainnet/grpc/vega/vega.proto#side
    side="SIDE_BUY",
    volume=1.0,
    price=100.0,
    # If this is true, function waits for confirmation and returns the order_id
    wait=False,
)

### Place a Market Order
self.vega.submit_market_order(
    trading_wallet=self.wallet_name,
    market_id=self.market_id,
    # https://docs.vega.xyz/docs/mainnet/grpc/vega/vega.proto#side
    side="SIDE_BUY",
    volume=1.0,
    # If this is true, function waits for confirmation and returns the order_id
    wait=False,
    # If this is true, order errors if the entire volume cannot be traded.
    # Otherwise, whatever can be traded will be.
    fill_or_kill=False,
)

### Amend an Order
self.vega.amend_order(
    trading_wallet=self.wallet_name,
    market_id=self.market_id,
    order_id=order_to_amend_id,
    price=10.1,
    volume_delta=10, # Note: Change in volume from existing order
)

### Cancel an Order
self.vega.cancel_order(
    self.wallet_name,
    self.market_id,
    order_id
)

### Retrieve Current Position & PnL (Profit & Loss)
positions = self.vega.positions_by_market(
    wallet_name=self.wallet_name, market_id=self.market_id
)
if len(positions) > 0:
    position = positions[0]
    volume = position.open_volume
    pnl = position.realised_pnl + position.unrealised_pnl
    entry_price = position.average_entry_price

### Retrieve Current Orders
orders = self.vega.orders_for_party_from_feed(
    self.wallet_name,
    self.market_id,
    live_only=True,  # set to False to also retrieve dead orders
)

### Retrieve Account Balance
curr_acct = self.party_account(
    wallet_name=wallet_name, asset_id=asset, market_id=None
)
general_balance = curr_acct.general  # The balance not currently in use by the market
margin_balance = curr_acct.margin  # The balance already being used as margin

#######################################################################
"""


from vega_sim.environment.agent import AgentWithWallet
from vega_sim.service import VegaService


class SimpleAgent(AgentWithWallet):
    def __init__(
        self,
        wallet_name: str,
        wallet_pass: str,
        market_name: str,
        asset_name: str,
        initial_balance: float,
    ):
        """Agent to extend with custom logic. A barebones skeleton which will
        currently faucet itself some tokens but then do nothing.

        Args:
            wallet_name:
                str, The name to use for this agent's wallet
            wallet_pass:
                str, The password which this agent uses to log in to the wallet
        """
        super().__init__(wallet_name=wallet_name, wallet_pass=wallet_pass)

        self.market_name = market_name
        self.asset_name = asset_name
        self.initial_balance = initial_balance

    def initialise(self, vega: VegaService):
        # Initialise wallet
        super().initialise(vega=vega, create_wallet=True)

        # Get market id
        self.market_id = self.vega.find_market_id(name=self.market_name)

        # Get asset id
        self.asset_id = self.vega.find_asset_id(symbol=self.asset_name)
        # Top up asset
        self.vega.mint(
            self.wallet_name,
            asset=self.asset_id,
            amount=self.initial_balance,
        )
        self.vega.wait_fn(2)

    def step(self, vega: VegaService):
        """This function is called once each loop through of the simulation.

        Here you want to put you main logic, what the trader will do at every
        time, perhaps placing an order.

        Args:
            vega:
                VegaService, a client with which to interact with the Vega
                network itself. Use this to query for the current state of
                the world
        """
        pass



import numpy as np
from collections import namedtuple
from typing import List, Optional

from vega_sim.scenario.ideal_market_maker.utils.strategy import A_S_MMmodel, GLFT_approx
from vega_sim.environment import VegaState
from vega_sim.environment.agent import StateAgentWithWallet
from vega_sim.null_service import VegaServiceNull
from vega_sim.proto.vega import vega as vega_protos

WalletConfig = namedtuple("WalletConfig", ["name", "passphrase"])

# Set up parties in the market/ Submit liquidity provision
MM_WALLET = WalletConfig("mm", "pin")

# Send selling/buying MOs to hit LP orders
TRADER_WALLET = WalletConfig("trader", "trader")

# Randomly posts LOs at buy/sell side to simulate real Market situation
RANDOM_WALLET = WalletConfig("random", "random")

# Pass opening auction
AUCTION1_WALLET = WalletConfig("AUCTION1", "AUCTION1pass")
AUCTION2_WALLET = WalletConfig("AUCTION2", "AUCTION2pass")

# Add another liquidity provider
LIQUIDITY = WalletConfig("liquidity", "liquiditypass")

# Terminate the market and send settlment price
TERMINATE_WALLET = WalletConfig("terminate", "terminate")

# informed trader wallet
INFORMED_WALLET = WalletConfig("INFORMED", "INFORMEDpass")


class OptimalMarketMaker(StateAgentWithWallet):
    def __init__(
        self,
        wallet_name: str,
        wallet_pass: str,
        terminate_wallet_name: str,
        terminate_wallet_pass: str,
        price_process: List[float],
        spread: float = 0.00002,
        num_steps: int = 180,
        market_order_arrival_rate: float = 5,
        pegged_order_fill_rate: float = 500,
        inventory_upper_boundary: int = 20,
        inventory_lower_boundary: int = -20,
        terminal_penalty_parameter: float = 10**-4,
        running_penalty_parameter: float = 5 * 10**-6,
        asset_decimal: int = 5,
        market_decimal: int = 5,
        market_position_decimal: int = 0,
        initial_asset_mint: float = 1e8,
        market_name: str = None,
        asset_name: str = None,
        commitamount: float = 100000,
        lp_fee: float = 0.001,
        settlement_price: Optional[float] = None,
        tag: str = "",
        random_state: Optional[np.random.RandomState] = None,
    ):
        super().__init__(wallet_name + tag, wallet_pass)
        self.terminate_wallet_name = terminate_wallet_name + tag
        self.terminate_wallet_pass = terminate_wallet_pass

        self.price_process = price_process
        self.spread = spread
        self.time = num_steps
        self.Lambda = market_order_arrival_rate
        self.kappa = pegged_order_fill_rate
        self.q_upper = inventory_upper_boundary
        self.q_lower = inventory_lower_boundary
        self.alpha = terminal_penalty_parameter
        self.phi = running_penalty_parameter
        self.adp = asset_decimal
        self.mdp = market_decimal
        self.market_position_decimal = market_position_decimal
        self.current_step = 0
        self.initial_asset_mint = initial_asset_mint
        self.commitamount = commitamount
        self.lp_fee = lp_fee
        self.settlement_price = (
            self.price_process[-1] if settlement_price is None else settlement_price
        )
        self.tag = tag
        self.market_name = f"ETH:USD_{self.tag}" if market_name is None else market_name
        self.asset_name = f"tDAI_{self.tag}" if asset_name is None else asset_name
        self.random_state = (
            random_state
            if random_state is not None
            else np.random.RandomState(seed=123)
        )

        self.long_horizon_estimate = num_steps >= 200

        if not self.long_horizon_estimate:
            self.optimal_bid, self.optimal_ask, _ = A_S_MMmodel(
                T=self.time / 60 / 24 / 365.25,
                dt=1 / 60 / 24 / 365.25,
                length=self.time + 1,
                mdp=self.mdp,
                q_upper=self.q_upper,
                q_lower=self.q_lower,
                kappa=self.kappa,
                Lambda=self.Lambda,
                alpha=self.alpha,
                phi=self.phi,
            )
        else:
            self.optimal_bid, self.optimal_ask = GLFT_approx(
                q_upper=self.q_upper,
                q_lower=self.q_lower,
                kappa=self.kappa,
                Lambda=self.Lambda,
                alpha=self.alpha,
                phi=self.phi,
            )

    def finalise(self):
        self.current_step += 1
        self.vega.settle_market(
            self.terminate_wallet_name, self.settlement_price, self.market_id
        )

    def initialise(self, vega: VegaServiceNull):
        # Initialise wallet for LP/ Settle Party
        super().initialise(vega=vega)
        self.vega.create_wallet(self.terminate_wallet_name, self.terminate_wallet_pass)

        # Faucet vega tokens
        self.vega.wait_for_datanode_sync()
        self.vega.mint(
            self.wallet_name,
            asset="VOTE",
            amount=1e4,
        )
        self.vega.wait_fn(5)

        # Create asset
        self.vega.create_asset(
            self.wallet_name,
            name=self.asset_name,
            symbol=self.asset_name,
            decimals=self.adp,
            max_faucet_amount=1e20,
        )
        self.vega.wait_fn(5)
        self.vega.wait_for_total_catchup()
        # Get asset id
        self.asset_id = self.vega.find_asset_id(symbol=self.asset_name)
        # Top up asset
        self.vega.mint(
            self.wallet_name,
            asset=self.asset_id,
            amount=self.initial_asset_mint,
        )
        self.vega.wait_for_total_catchup()

        # Set up a future market
        self.vega.create_simple_market(
            market_name=self.market_name,
            proposal_wallet=self.wallet_name,
            settlement_asset_id=self.asset_id,
            termination_wallet=self.terminate_wallet_name,
            market_decimals=self.mdp,
            position_decimals=self.market_position_decimal,
            future_asset=self.asset_name,
        )
        self.vega.wait_for_total_catchup()

        # Get market id
        self.market_id = self.vega.find_market_id(name=self.market_name)

        vega.submit_liquidity(
            wallet_name=self.wallet_name,
            market_id=self.market_id,
            commitment_amount=self.commitamount,
            fee=self.lp_fee,
            buy_specs=[("PEGGED_REFERENCE_MID", 0.1, 1)],
            sell_specs=[("PEGGED_REFERENCE_MID", 0.1, 1)],
            is_amendment=False,
        )
        self.bid_depth, self.ask_depth = self.OptimalStrategy(0)

    def num_MarketOrders(self):
        num_buyMO = self.random_state.poisson(self.Lambda)
        num_sellMO = self.random_state.poisson(self.Lambda)
        return num_buyMO, num_sellMO

    def num_LimitOrderHit(self, bid_depth, ask_depth, num_buyMO, num_sellMO):
        poss_bid_filled = np.exp(-self.kappa * bid_depth)
        poss_ask_filled = np.exp(-self.kappa * ask_depth)

        num_BidLimitOrderHit = num_sellMO - self.random_state.binomial(
            num_sellMO, poss_bid_filled
        )
        num_AskLimitOrderHit = num_buyMO - self.random_state.binomial(
            num_buyMO, poss_ask_filled
        )

        # If the numebr of LOs at ask side is 0
        if num_AskLimitOrderHit == 0:
            # In this case, the MOs still cannot hit the LP orders
            num_AskLimitOrderHit += 1
            num_buyMO += 1
        if num_BidLimitOrderHit == 0:
            num_BidLimitOrderHit += 1
            num_sellMO += 1

        return num_BidLimitOrderHit, num_AskLimitOrderHit

    def OptimalStrategy(self, current_position):
        if current_position >= self.q_upper:
            current_bid_depth = (
                self.optimal_bid[self.current_step, 0]
                if not self.long_horizon_estimate
                else self.optimal_bid[0]
            )
            current_ask_depth = 1 / 10**self.mdp
        elif current_position <= self.q_lower:
            current_bid_depth = 1 / 10**self.mdp
            current_ask_depth = (
                self.optimal_ask[self.current_step, -1]
                if not self.long_horizon_estimate
                else self.optimal_ask[-1]
            )
        else:
            current_bid_depth = (
                self.optimal_bid[
                    self.current_step, int(self.q_upper - 1 - current_position)
                ]
                if not self.long_horizon_estimate
                else self.optimal_bid[int(self.q_upper - 1 - current_position)]
            )
            current_ask_depth = (
                self.optimal_ask[
                    self.current_step, int(self.q_upper - current_position)
                ]
                if not self.long_horizon_estimate
                else self.optimal_ask[int(self.q_upper - current_position)]
            )

        return current_bid_depth, current_ask_depth

    def AvoidCrossedOrder(self):
        new_price = self.price_process[self.current_step]
        price = (
            self.price_process[self.current_step - 1]
            if self.current_step > 0
            else self.price_process[self.current_step]
        )

        if min(self.bid_depth, self.ask_depth) > np.abs(new_price - price) / 2:
            pass

        else:
            temp_depth = round(
                np.abs(new_price - price) / 2 + 5 * self.spread,
                self.mdp,
            )
            self.vega.submit_simple_liquidity(
                wallet_name=self.wallet_name,
                market_id=self.market_id,
                commitment_amount=self.commitamount,
                fee=self.lp_fee,
                reference_buy="PEGGED_REFERENCE_MID",
                reference_sell="PEGGED_REFERENCE_MID",
                delta_buy=temp_depth,
                delta_sell=temp_depth,
                is_amendment=True,
            )

    def step(self, vega_state: VegaState):
        # Each step, MM posts optimal bid/ask depths
        position = self.vega.positions_by_market(
            wallet_name=self.wallet_name, market_id=self.market_id
        )

        current_position = int(position[0].open_volume) if position else 0
        self.bid_depth, self.ask_depth = self.OptimalStrategy(current_position)

        self.num_buyMO, self.num_sellMO = self.num_MarketOrders()
        self.num_bidhit, self.num_askhit = self.num_LimitOrderHit(
            bid_depth=self.bid_depth,
            ask_depth=self.ask_depth,
            num_buyMO=self.num_buyMO,
            num_sellMO=self.num_sellMO,
        )

        self.vega.submit_simple_liquidity(
            wallet_name=self.wallet_name,
            market_id=self.market_id,
            commitment_amount=self.commitamount,
            fee=self.lp_fee,
            reference_buy="PEGGED_REFERENCE_MID",
            reference_sell="PEGGED_REFERENCE_MID",
            delta_buy=self.bid_depth,
            delta_sell=self.ask_depth,
            is_amendment=True,
        )
        self.current_step += 1


class MarketOrderTrader(StateAgentWithWallet):
    def __init__(
        self,
        wallet_name: str,
        wallet_pass: str,
        market_name: str = None,
        asset_name: str = None,
        initial_asset_mint: float = 1e8,
        num_buy_market_order: int = None,
        num_sell_market_order: int = None,
        tag: str = "",
    ):
        super().__init__(wallet_name + str(tag), wallet_pass)
        self.initial_asset_mint = initial_asset_mint
        self.num_buyMO = num_buy_market_order
        self.num_sellMO = num_sell_market_order
        self.tag = tag
        self.market_name = f"ETH:USD_{self.tag}" if market_name is None else market_name
        self.asset_name = f"tDAI_{self.tag}" if asset_name is None else asset_name

    def initialise(self, vega: VegaServiceNull):
        # Initialise wallet
        super().initialise(vega=vega)

        # Get market id
        self.market_id = self.vega.find_market_id(name=self.market_name)

        # Get asset id
        tDAI_id = self.vega.find_asset_id(symbol=self.asset_name)
        # Top up asset
        self.vega.mint(
            self.wallet_name,
            asset=tDAI_id,
            amount=self.initial_asset_mint,
        )
        self.vega.wait_fn(2)

    def step_buy(self, vega_state: VegaState):
        if self.num_buyMO > 0:
            self.vega.submit_market_order(
                trading_wallet=self.wallet_name,
                market_id=self.market_id,
                side="SIDE_BUY",
                volume=self.num_buyMO,
                wait=False,
                fill_or_kill=False,
            )

    def step_sell(self, vega_state: VegaState):
        if self.num_sellMO > 0:
            self.vega.submit_market_order(
                trading_wallet=self.wallet_name,
                market_id=self.market_id,
                side="SIDE_SELL",
                volume=self.num_sellMO,
                wait=False,
                fill_or_kill=False,
            )


class LimitOrderTrader(StateAgentWithWallet):
    def __init__(
        self,
        wallet_name: str,
        wallet_pass: str,
        price_process: List[float],
        num_post_at_bid: int = None,
        num_post_at_ask: int = None,
        spread: float = 0.00002,
        initial_price: float = 0.3,
        market_decimal: int = 5,
        asset_decimal: int = 5,
        initial_asset_mint: float = 1e8,
        market_name: str = None,
        asset_name: str = None,
        tag: str = "",
    ):
        super().__init__(wallet_name + tag, wallet_pass)
        self.num_post_at_bid = num_post_at_bid
        self.num_post_at_ask = num_post_at_ask
        self.price_process = price_process
        self.spread = spread
        self.initial_price = initial_price
        self.mdp = market_decimal
        self.adp = asset_decimal
        self.initial_asset_mint = initial_asset_mint
        self.current_step = 0
        self.tag = tag
        self.market_name = f"ETH:USD_{self.tag}" if market_name is None else market_name
        self.asset_name = f"tDAI_{self.tag}" if asset_name is None else asset_name

    def initialise(self, vega: VegaServiceNull):
        # Initialise wallet
        super().initialise(vega=vega)
        # Get market id
        self.market_id = self.vega.find_market_id(name=self.market_name)

        # Get asset id
        tDAI_id = self.vega.find_asset_id(symbol=self.asset_name)
        # Top up asset
        self.vega.mint(
            self.wallet_name,
            asset=tDAI_id,
            amount=self.initial_asset_mint,
        )
        self.vega.wait_fn(2)

        self.buy_order_id = self.vega.submit_order(
            trading_wallet=self.wallet_name,
            market_id=self.market_id,
            order_type="TYPE_LIMIT",
            time_in_force="TIME_IN_FORCE_GTC",
            side="SIDE_BUY",
            volume=1,
            price=round(self.initial_price - self.spread, self.mdp),
            wait=True,
        )

        self.sell_order_id = self.vega.submit_order(
            trading_wallet=self.wallet_name,
            market_id=self.market_id,
            order_type="TYPE_LIMIT",
            time_in_force="TIME_IN_FORCE_GTC",
            side="SIDE_SELL",
            volume=1,
            price=round(self.initial_price + self.spread, self.mdp),
            wait=True,
        )

    def step_amendprice(self, vega_state: VegaState):
        new_price = self.price_process[self.current_step]
        price = (
            self.price_process[self.current_step - 1]
            if self.current_step > 0
            else self.price_process[self.current_step]
        )

        first_side = (
            vega_protos.SIDE_BUY if new_price < price else vega_protos.SIDE_SELL
        )

        if first_side == vega_protos.SIDE_BUY:
            self.vega.amend_order(
                trading_wallet=self.wallet_name,
                market_id=self.market_id,
                order_id=self.buy_order_id,
                price=round(
                    new_price - self.spread / 2,
                    self.mdp,
                ),
            )

        self.vega.amend_order(
            trading_wallet=self.wallet_name,
            market_id=self.market_id,
            order_id=self.sell_order_id,
            price=round(
                new_price + self.spread / 2,
                self.mdp,
            ),
        )

        if first_side == vega_protos.SIDE_SELL:
            self.vega.amend_order(
                trading_wallet=self.wallet_name,
                market_id=self.market_id,
                order_id=self.buy_order_id,
                price=round(
                    new_price - self.spread / 2,
                    self.mdp,
                ),
            )

    def step_limitorders(self, vega_state: VegaState):
        if self.num_post_at_bid > 1:
            random_delta = self.spread / 2

            self.vega.submit_order(
                trading_wallet=self.wallet_name,
                market_id=self.market_id,
                time_in_force="TIME_IN_FORCE_GTC",
                order_type="TYPE_LIMIT",
                side="SIDE_BUY",
                volume=self.num_post_at_bid - 1,
                price=self.price_process[self.current_step] - random_delta,
                wait=False,
            )

        if self.num_post_at_ask > 1:
            random_delta = self.spread / 2

            self.vega.submit_order(
                trading_wallet=self.wallet_name,
                market_id=self.market_id,
                time_in_force="TIME_IN_FORCE_GTC",
                order_type="TYPE_LIMIT",
                side="SIDE_SELL",
                volume=self.num_post_at_ask - 1,
                price=self.price_process[self.current_step] + random_delta,
                wait=False,
            )

    def step_limitorderask(self, vega_state: VegaState):
        self.sell_order_id = self.vega.submit_order(
            trading_wallet=self.wallet_name,
            market_id=self.market_id,
            order_type="TYPE_LIMIT",
            time_in_force="TIME_IN_FORCE_GTC",
            side="SIDE_SELL",
            volume=1,
            price=round(self.price_process[self.current_step] + self.spread, self.mdp),
            wait=True,
        )

    def step_limitorderbid(self, vega_state: VegaState):
        self.buy_order_id = self.vega.submit_order(
            trading_wallet=self.wallet_name,
            market_id=self.market_id,
            order_type="TYPE_LIMIT",
            time_in_force="TIME_IN_FORCE_GTC",
            side="SIDE_BUY",
            volume=1,
            price=round(self.price_process[self.current_step] - self.spread, self.mdp),
            wait=True,
        )

        self.current_step += 1


class OpenAuctionPass(StateAgentWithWallet):
    def __init__(
        self,
        wallet_name: str,
        wallet_pass: str,
        side: str,
        market_name: str = None,
        asset_name: str = None,
        initial_asset_mint: float = 1e8,
        initial_price: float = 0.3,
        tag: str = "",
    ):
        super().__init__(wallet_name + tag, wallet_pass)
        self.side = side
        self.initial_price = initial_price
        self.initial_asset_mint = initial_asset_mint
        self.tag = tag
        self.market_name = f"ETH:USD_{self.tag}" if market_name is None else market_name
        self.asset_name = f"tDAI_{self.tag}" if asset_name is None else asset_name

    def initialise(self, vega: VegaServiceNull):
        # Initialise wallet
        super().initialise(vega=vega)
        # Get market id
        self.market_id = self.vega.find_market_id(name=self.market_name)

        # Get asset id
        tDAI_id = self.vega.find_asset_id(symbol=self.asset_name)
        # Top up asset
        self.vega.mint(
            self.wallet_name,
            asset=tDAI_id,
            amount=self.initial_asset_mint,
        )
        self.vega.wait_fn(2)

        self.vega.submit_order(
            trading_wallet=self.wallet_name,
            market_id=self.market_id,
            order_type="TYPE_LIMIT",
            time_in_force="TIME_IN_FORCE_GTC",
            side=self.side,
            volume=1,
            price=self.initial_price,
        )

    def step(self, vega_state: VegaState):
        pass


class OptimalLiquidityProvider(StateAgentWithWallet):
    def __init__(
        self,
        wallet_name: str,
        wallet_pass: str,
        num_steps: int = 100,
        market_order_arrival_rate: float = 10,
        pegged_order_fill_rate: float = 500,
        inventory_upper_boundary: int = 20,
        inventory_lower_boundary: int = -20,
        terminal_penalty_parameter: float = 10**-4,
        running_penalty_parameter: float = 5 * 10**-6,
        initial_asset_mint: float = 1e8,
        market_name: str = None,
        asset_name: str = None,
        entry_step: int = 0,
        commitamount: float = 100000,
        lp_fee: float = 0.001,
        tag: str = "",
    ):
        super().__init__(wallet_name + tag, wallet_pass)
        self.current_step = 0
        self.initial_asset_mint = initial_asset_mint
        self.entry_step = entry_step
        self.commitamount = commitamount
        self.lp_fee = lp_fee
        self.time = num_steps
        self.Lambda = market_order_arrival_rate
        self.kappa = pegged_order_fill_rate
        self.q_upper = inventory_upper_boundary
        self.q_lower = inventory_lower_boundary
        self.alpha = terminal_penalty_parameter
        self.phi = running_penalty_parameter
        self.tag = tag
        self.market_name = f"ETH:USD_{self.tag}" if market_name is None else market_name
        self.asset_name = f"tDAI_{self.tag}" if asset_name is None else asset_name
        self.long_horizon_estimate = num_steps >= 200

    def initialise(self, vega: VegaServiceNull):
        # Initialise wallet for LP/ Settle Party
        super().initialise(vega=vega)

        # Get market id
        self.market_id = self.vega.find_market_id(name=self.market_name)

        # Get asset id
        tDAI_id = self.vega.find_asset_id(symbol=self.asset_name)
        # Top up asset
        self.vega.mint(
            self.wallet_name,
            asset=tDAI_id,
            amount=self.initial_asset_mint,
        )
        # Get asset id
        self.asset_id = self.vega.find_asset_id(symbol=self.asset_name)
        # Top up asset
        self.vega.mint(
            self.wallet_name,
            asset=self.asset_id,
            amount=self.initial_asset_mint,
        )
        self.vega.wait_for_total_catchup()

        # Get market decimal place/asset decimal place
        self.mdp = list(self.vega._market_price_decimals.values())[0]
        self.adp = list(self.vega._asset_decimals.values())[0]

        # Get optimal market making strategy
        if not self.long_horizon_estimate:
            self.optimal_bid, self.optimal_ask, _ = A_S_MMmodel(
                T=self.time / 60 / 24 / 365.25,
                dt=1 / 60 / 24 / 365.25,
                length=self.time + 1,
                mdp=self.mdp,
                q_upper=self.q_upper,
                q_lower=self.q_lower,
                kappa=self.kappa,
                Lambda=self.Lambda,
                alpha=self.alpha,
                phi=self.phi,
            )
        else:
            self.optimal_bid, self.optimal_ask = GLFT_approx(
                q_upper=self.q_upper,
                q_lower=self.q_lower,
                kappa=self.kappa,
                Lambda=self.Lambda,
                alpha=self.alpha,
                phi=self.phi,
            )

    def OptimalStrategy(self, current_position):
        if current_position >= self.q_upper:
            current_bid_depth = (
                self.optimal_bid[self.current_step, 0]
                if not self.long_horizon_estimate
                else self.optimal_bid[0]
            )
            current_ask_depth = 1 / 10**self.mdp
        elif current_position <= self.q_lower:
            current_bid_depth = 1 / 10**self.mdp
            current_ask_depth = (
                self.optimal_ask[self.current_step, -1]
                if not self.long_horizon_estimate
                else self.optimal_ask[-1]
            )
        else:
            current_bid_depth = (
                self.optimal_bid[
                    self.current_step, int(self.q_upper - 1 - current_position)
                ]
                if not self.long_horizon_estimate
                else self.optimal_bid[int(self.q_upper - 1 - current_position)]
            )
            current_ask_depth = (
                self.optimal_ask[
                    self.current_step, int(self.q_upper - current_position)
                ]
                if not self.long_horizon_estimate
                else self.optimal_ask[int(self.q_upper - current_position)]
            )

        return current_bid_depth, current_ask_depth

    def step(self, vega_state: VegaState):
        if self.current_step < self.entry_step:
            self.current_step += 1
            return

        elif self.current_step == self.entry_step:
            current_position = 0
            self.bid_depth, self.ask_depth = self.OptimalStrategy(current_position)
            self.vega.submit_simple_liquidity(
                wallet_name=self.wallet_name,
                market_id=self.market_id,
                commitment_amount=self.commitamount,
                fee=self.lp_fee,
                reference_buy="PEGGED_REFERENCE_MID",
                reference_sell="PEGGED_REFERENCE_MID",
                delta_buy=self.bid_depth,
                delta_sell=self.ask_depth,
                is_amendment=False,
            )
            self.current_step += 1
            return

        position = self.vega.positions_by_market(
            wallet_name=self.wallet_name, market_id=self.market_id
        )

        current_position = int(position[0].open_volume) if position else 0
        self.bid_depth, self.ask_depth = self.OptimalStrategy(current_position)

        self.vega.submit_simple_liquidity(
            wallet_name=self.wallet_name,
            market_id=self.market_id,
            commitment_amount=self.commitamount,
            fee=self.lp_fee,
            reference_buy="PEGGED_REFERENCE_MID",
            reference_sell="PEGGED_REFERENCE_MID",
            delta_buy=self.bid_depth,
            delta_sell=self.ask_depth,
            is_amendment=True,
        )

        self.current_step += 1


class InformedTrader(StateAgentWithWallet):
    def __init__(
        self,
        wallet_name: str,
        wallet_pass: str,
        price_process: List[float],
        market_name: str = None,
        asset_name: str = None,
        initial_asset_mint: float = 1e8,
        proportion_taken: float = 0.8,
        tag: str = "",
    ):
        super().__init__(wallet_name + str(tag), wallet_pass)
        self.initial_asset_mint = initial_asset_mint
        self.price_process = price_process
        self.current_step = 0
        self.sim_length = len(price_process)
        self.tag = tag
        self.proportion_taken = proportion_taken
        self.market_name = f"ETH:USD_{self.tag}" if market_name is None else market_name
        self.asset_name = f"tDAI_{self.tag}" if asset_name is None else asset_name

    def initialise(self, vega: VegaServiceNull):
        # Initialise wallet
        super().initialise(vega=vega)

        # Get market id
        self.market_id = self.vega.find_market_id(name=self.market_name)

        # Get asset id
        tDAI_id = self.vega.find_asset_id(symbol=self.asset_name)
        # Top up asset
        self.vega.mint(
            self.wallet_name,
            asset=tDAI_id,
            amount=self.initial_asset_mint,
        )

        self.pdp = self.vega._market_pos_decimals.get(self.market_id, {})
        self.vega.wait_for_total_catchup()

    def step(self, vega_state: VegaState):
        position = self.vega.positions_by_market(
            wallet_name=self.wallet_name, market_id=self.market_id
        )
        current_position = int(position[0].open_volume) if position else 0
        trade_side = (
            vega_protos.vega.Side.SIDE_BUY
            if current_position < 0
            else vega_protos.vega.Side.SIDE_SELL
        )
        if current_position:
            self.vega.submit_market_order(
                trading_wallet=self.wallet_name,
                market_id=self.market_id,
                side=trade_side,
                volume=current_position,
                wait=True,
                fill_or_kill=False,
            )

        order_book = self.vega.market_depth(market_id=self.market_id)

        price = self.price_process[self.current_step]
        next_price = self.price_process[self.current_step + 1]

        trade_side = (
            vega_protos.SIDE_BUY if price < next_price else vega_protos.SIDE_SELL
        )

        if price < next_price:
            volume = sum(
                [order.volume for order in order_book.sells if order.price < next_price]
            )
        else:
            volume = sum(
                [order.volume for order in order_book.buys if order.price > next_price]
            )

        volume = round(self.proportion_taken * volume, self.pdp)

        if volume:
            self.vega.submit_market_order(
                trading_wallet=self.wallet_name,
                market_id=self.market_id,
                side=trade_side,
                volume=volume,
                wait=False,
                fill_or_kill=False,
            )



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from blockexplorer import blockexplorer_pb2 as blockexplorer_dot_blockexplorer__pb2


class BlockExplorerServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetTransaction = channel.unary_unary(
            "/blockexplorer.api.v1.BlockExplorerService/GetTransaction",
            request_serializer=blockexplorer_dot_blockexplorer__pb2.GetTransactionRequest.SerializeToString,
            response_deserializer=blockexplorer_dot_blockexplorer__pb2.GetTransactionResponse.FromString,
        )
        self.ListTransactions = channel.unary_unary(
            "/blockexplorer.api.v1.BlockExplorerService/ListTransactions",
            request_serializer=blockexplorer_dot_blockexplorer__pb2.ListTransactionsRequest.SerializeToString,
            response_deserializer=blockexplorer_dot_blockexplorer__pb2.ListTransactionsResponse.FromString,
        )


class BlockExplorerServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def GetTransaction(self, request, context):
        """Get transaction

        Get a transaction from the Vega blockchain
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ListTransactions(self, request, context):
        """List transactions

        List transactions from the Vega blockchain
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_BlockExplorerServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "GetTransaction": grpc.unary_unary_rpc_method_handler(
            servicer.GetTransaction,
            request_deserializer=blockexplorer_dot_blockexplorer__pb2.GetTransactionRequest.FromString,
            response_serializer=blockexplorer_dot_blockexplorer__pb2.GetTransactionResponse.SerializeToString,
        ),
        "ListTransactions": grpc.unary_unary_rpc_method_handler(
            servicer.ListTransactions,
            request_deserializer=blockexplorer_dot_blockexplorer__pb2.ListTransactionsRequest.FromString,
            response_serializer=blockexplorer_dot_blockexplorer__pb2.ListTransactionsResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "blockexplorer.api.v1.BlockExplorerService", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class BlockExplorerService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def GetTransaction(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/blockexplorer.api.v1.BlockExplorerService/GetTransaction",
            blockexplorer_dot_blockexplorer__pb2.GetTransactionRequest.SerializeToString,
            blockexplorer_dot_blockexplorer__pb2.GetTransactionResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ListTransactions(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/blockexplorer.api.v1.BlockExplorerService/ListTransactions",
            blockexplorer_dot_blockexplorer__pb2.ListTransactionsRequest.SerializeToString,
            blockexplorer_dot_blockexplorer__pb2.ListTransactionsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



import os
import csv
from pathlib import Path


def createfile_MMmodel(folder_path, strategy):
    """
    Create a csv file to load the output data from current strategy.
    If the file exists, clear the file; If not, create one.
    """

    Path(folder_path).mkdir(exist_ok=True)
    file_path = folder_path + strategy
    if os.path.isfile(file_path):
        f = open(file_path, "w")
        f.truncate()
        f.close()

    header = [
        "Iteration",
        "Time Step",
        "LP: General Account",
        "LP: Margin Account",
        "LP: Bond Account",
        "LP: General Pnl",
        "LP: RealisedPnl",
        "LP: UnrealisedPnl",
        "LP: Position",
        "LP: Bid Depth",
        "LP: Ask Depth",
        "Buying MOs",
        "Selling MOs",
        "LOs at bid",
        "LOs at ask",
        "External Midprice",
        "Markprice",
        "Average Entry price",
        "InsurancePool",
        "LiquifeeAccount",
        "InfrafeeAccount",
        "Market Trading mode",
        "Market Sate",
    ]

    with open(file_path, "a", encoding="UTF8", newline="") as f:
        writer = csv.writer(f)
        writer.writerow(header)
    return file_path


def createfile_multiLPs(
    strategy="MC_multiLPs_MMmodel.csv",
    num_lps=10,
):
    """
    Create a csv file to load the output data from current strategy.
    If the file exists, clear the file; If not, create one.
    """
    path = os.getcwd() + "/"
    file = path + strategy
    if os.path.isfile(file):
        f = open(file, "w")
        f.truncate()
        f.close()

    header = [
        "Iteration",
        "Time Step",
    ]

    header_GA = ["LP{}: General".format(i + 1) for i in range(num_lps)]
    header_MA = ["LP{}: Margin".format(i + 1) for i in range(num_lps)]
    header_BA = ["LP{}: Bond".format(i + 1) for i in range(num_lps)]
    header_PNL = ["LP{}: GeneralPnl".format(i + 1) for i in range(num_lps)]
    header_Re = ["LP{}: RealisedPnl".format(i + 1) for i in range(num_lps)]
    header_Un = ["LP{}: UnrealisedPnl".format(i + 1) for i in range(num_lps)]
    header_Pos = ["LP{}: Position".format(i + 1) for i in range(num_lps)]
    header_bid = ["LP{}: Bid".format(i + 1) for i in range(num_lps)]
    header_ask = ["LP{}: Ask".format(i + 1) for i in range(num_lps)]
    header_entry = ["LP{}: Entryprice".format(i + 1) for i in range(num_lps)]
    header_els = ["LP{}: ELS".format(i + 1) for i in range(num_lps)]

    header = (
        header
        + header_GA
        + header_MA
        + header_BA
        + header_PNL
        + header_Re
        + header_Un
        + header_Pos
        + header_bid
        + header_ask
        + header_entry
        + header_els
    )

    header += [
        "Buying MOs",
        "Selling MOs",
        "LOs at bid",
        "LOs at ask",
        "External Midprice",
        "Markprice",
        "InsurancePool",
        "LiquifeeAccount",
        "InfrafeeAccount",
        "Market Trading mode",
        "Market Sate",
    ]

    with open(file, "a", encoding="UTF8", newline="") as f:
        writer = csv.writer(f)
        writer.writerow(header)


def createfile_infotrader(
    strategy,
):
    """
    Create a csv file to load the output data from current strategy.
    If the file exists, clear the file; If not, create one.
    """
    path = os.getcwd() + "/"
    file = path + strategy
    if os.path.isfile(file):
        f = open(file, "w")
        f.truncate()
        f.close()

    header = [
        "Iteration",
        "Time Step",
        "LP: General Account",
        "LP: Margin Account",
        "LP: Bond Account",
        "LP: GeneralPnl",
        "LP: RealisedPnl",
        "LP: UnrealisedPnl",
        "LP: Position",
        "Trader: General Account",
        "Trader: Margin Account",
        "Trader: GeneralPnl",
        "Trader: RealisedPnl",
        "Trader: UnrealisedPnl",
        "Trader: Position",
        "LP: Bid Depth",
        "LP: Ask Depth",
        "Total buyingMO",
        "Total sellingMO",
        "BuyingMO Trader",
        "SellingMO Trader",
        "LOs at bid",
        "LOs at ask",
        "External Midprice",
        "Markprice",
        "Average Entry price",
        "InsurancePool",
        "LiquifeeAccount",
        "InfrafeeAccount",
        "Market Trading mode",
        "Market Sate",
    ]

    with open(file, "a", encoding="UTF8", newline="") as f:
        writer = csv.writer(f)
        writer.writerow(header)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: vega/commands/v1/signature.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n vega/commands/v1/signature.proto\x12\x10vega.commands.v1"O\n\tSignature\x12\x14\n\x05value\x18\x01 \x01(\tR\x05value\x12\x12\n\x04\x61lgo\x18\x02 \x01(\tR\x04\x61lgo\x12\x18\n\x07version\x18\x03 \x01(\rR\x07versionB3Z1code.vegaprotocol.io/vega/protos/vega/commands/v1b\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "vega.commands.v1.signature_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = (
        b"Z1code.vegaprotocol.io/vega/protos/vega/commands/v1"
    )
    _SIGNATURE._serialized_start = 54
    _SIGNATURE._serialized_end = 133
# @@protoc_insertion_point(module_scope)



import grpc
from abc import ABC
from vega_sim.proto.data_node.api.v1 import trading_data_grpc
from vega_sim.proto.data_node.api.v2 import trading_data_grpc as trading_data_grpc_v2
from vega_sim.proto.vega.api.v1 import core_grpc, corestate_grpc


class GRPCClient(ABC):
    """
    A `GRPCClient` talks to a gRPC endpoint.
    """

    STUB_CLASS = None

    def __init__(self, url: str, channel=None) -> None:
        if url is None:
            raise Exception("Missing node URL")
        self.url = url

        if channel is None:
            # get a gRPC channel
            channel = grpc.insecure_channel(self.url)
            grpc.channel_ready_future(channel).result(timeout=10)

        self.channel = channel
        self._client = self.STUB_CLASS(self.channel)

    def __getattr__(self, funcname):
        return getattr(self._client, funcname)


class VegaTradingDataClient(GRPCClient):
    """
    The Vega Trading Data Client talks to a back-end node.
    """

    STUB_CLASS = trading_data_grpc.TradingDataServiceStub


class VegaTradingDataClientV2(GRPCClient):
    """
    The Vega Trading Data Client talks to a back-end node.
    """

    STUB_CLASS = trading_data_grpc_v2.TradingDataServiceStub


class VegaCoreClient(GRPCClient):
    """
    The Vega Core Client talks to a back-end node.
    """

    STUB_CLASS = core_grpc.CoreServiceStub


class VegaCoreStateClient(GRPCClient):
    """
    The Vega Core Client talks to a back-end node.
    """

    STUB_CLASS = corestate_grpc.CoreStateServiceStub



from __future__ import annotations

import atexit
import functools
from io import BufferedWriter
import logging
import multiprocessing
import os
import shutil
import signal
import socket
import subprocess
import tempfile
import time
import webbrowser
from collections import namedtuple
from contextlib import closing
from enum import Enum, auto
from os import path
from typing import Dict, List, Optional, Set

import requests
import toml
from urllib3.exceptions import MaxRetryError

from vega_sim import vega_bin_path, vega_home_path
from vega_sim.service import VegaService
from vega_sim.wallet.base import Wallet
from vega_sim.wallet.slim_wallet import (
    SlimWallet,
)
from vega_sim.wallet.vega_wallet import VegaWallet

logger = logging.getLogger(__name__)

PortUpdateConfig = namedtuple(
    "PortUpdateConfig", ["file_path", "config_path", "key", "val_func"]
)


class Ports(Enum):
    DATA_NODE_GRPC = auto()
    DATA_NODE_REST = auto()
    DATA_NODE_GRAPHQL = auto()
    DATA_NODE_POSTGRES = auto()
    FAUCET = auto()
    WALLET = auto()
    VEGA_NODE = auto()
    CORE_GRPC = auto()
    CORE_REST = auto()
    BROKER = auto()
    METRICS = auto()
    PPROF = auto()
    CONSOLE = auto()


PORT_UPDATERS = {
    Ports.DATA_NODE_GRPC: [
        PortUpdateConfig(
            ("config", "data-node", "config.toml"),
            ["API"],
            "Port",
            lambda port: port,
        ),
        PortUpdateConfig(
            ("config", "data-node", "config.toml"),
            ["Gateway", "Node"],
            "Port",
            lambda port: port,
        ),
        PortUpdateConfig(
            ("config", "wallet-service", "networks", "local.toml"),
            ["API", "GRPC"],
            "Hosts",
            lambda port: [f"localhost:{port}"],
        ),
    ],
    Ports.DATA_NODE_REST: [
        PortUpdateConfig(
            ("config", "data-node", "config.toml"),
            ["Gateway", "REST"],
            "Port",
            lambda port: port,
        ),
        PortUpdateConfig(
            ("config", "wallet-service", "networks", "local.toml"),
            ["API", "REST"],
            "Hosts",
            lambda port: [f"localhost:{port}"],
        ),
    ],
    Ports.DATA_NODE_POSTGRES: [
        PortUpdateConfig(
            ("config", "data-node", "config.toml"),
            ["SQLStore", "ConnectionConfig"],
            "Port",
            lambda port: port,
        ),
    ],
    Ports.FAUCET: [
        PortUpdateConfig(
            ("config", "faucet", "config.toml"), [], "Port", lambda port: port
        ),
    ],
    Ports.WALLET: [
        PortUpdateConfig(
            ("config", "wallet-service", "networks", "local.toml"),
            [],
            "Port",
            lambda port: port,
        ),
    ],
    Ports.VEGA_NODE: [
        PortUpdateConfig(
            ("config", "node", "config.toml"),
            ["Blockchain", "Null"],
            "Port",
            lambda port: port,
        ),
    ],
    Ports.DATA_NODE_GRAPHQL: [
        PortUpdateConfig(
            ("config", "data-node", "config.toml"),
            ["Gateway", "GraphQL"],
            "Port",
            lambda port: port,
        ),
        PortUpdateConfig(
            ("config", "wallet-service", "networks", "local.toml"),
            ["API", "GraphQL"],
            "Hosts",
            lambda port: [f"localhost:{port}"],
        ),
    ],
    Ports.CORE_GRPC: [
        PortUpdateConfig(
            ("config", "faucet", "config.toml"),
            ["Node"],
            "Port",
            lambda port: port,
        ),
        PortUpdateConfig(
            ("config", "node", "config.toml"),
            ["API"],
            "Port",
            lambda port: port,
        ),
        PortUpdateConfig(
            ("config", "data-node", "config.toml"),
            ["API"],
            "CoreNodeGRPCPort",
            lambda port: port,
        ),
    ],
    Ports.CORE_REST: [
        PortUpdateConfig(
            ("config", "node", "config.toml"),
            ["API", "REST"],
            "Port",
            lambda port: port,
        ),
    ],
    Ports.BROKER: [
        PortUpdateConfig(
            ("config", "data-node", "config.toml"),
            ["Broker", "SocketConfig"],
            "Port",
            lambda port: port,
        ),
        PortUpdateConfig(
            ("config", "node", "config.toml"),
            ["Broker", "Socket"],
            "Port",
            lambda port: port,
        ),
    ],
    Ports.METRICS: [
        PortUpdateConfig(
            ("config", "data-node", "config.toml"),
            ["Metrics"],
            "Port",
            lambda port: port,
        ),
        PortUpdateConfig(
            ("config", "node", "config.toml"),
            ["Metrics"],
            "Port",
            lambda port: port,
        ),
    ],
    Ports.PPROF: [
        PortUpdateConfig(
            ("config", "data-node", "config.toml"),
            ["Pprof"],
            "Port",
            lambda port: port,
        ),
        PortUpdateConfig(
            ("config", "node", "config.toml"),
            ["Pprof"],
            "Port",
            lambda port: port,
        ),
    ],
    Ports.CONSOLE: [
        PortUpdateConfig(
            ("config", "wallet-service", "networks", "local.toml"),
            ["Console"],
            "LocalPort",
            lambda port: port,
        ),
    ],
}


class VegaStartupTimeoutError(Exception):
    pass


class ServiceNotStartedError(Exception):
    pass


class SocketNotFoundError(Exception):
    pass


def find_free_port(existing_set: Optional[Set[int]] = None):
    ret_sock = 0
    existing_set = (
        existing_set.union(set([ret_sock]))
        if existing_set is not None
        else set([ret_sock])
    )

    num_tries = 0
    while ret_sock in existing_set:
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
            s.bind(("", 0))
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            ret_sock = s.getsockname()[1]

        num_tries += 1
        if num_tries >= 100:
            # Arbitrary high number. If we try 100 times and fail to find
            # a port it seems reasonable to give up
            raise SocketNotFoundError("Failed finding a free socket")

    return ret_sock


def _terminate_proc(
    proc: subprocess.Popen[bytes], out_file: BufferedWriter, err_file: BufferedWriter
) -> None:
    proc.terminate()
    out_file.close()
    err_file.close()


def _popen_process(
    popen_args: List[str],
    dir_root: str,
    log_name: str,
    env: Optional[Dict[str, str]] = None,
) -> subprocess.Popen[bytes]:
    out = open(path.join(dir_root, f"{log_name}.out"), "wb")
    err = open(path.join(dir_root, f"{log_name}.err"), "wb")
    sub_proc = subprocess.Popen(
        popen_args, stdout=out, stderr=err, env=env, close_fds=True
    )

    atexit.register(functools.partial(_terminate_proc, sub_proc, out, err))
    return sub_proc


def _update_node_config(
    vega_home: str,
    port_config: Dict[Ports, int],
    transactions_per_block: int = 1,
    block_duration: str = "1s",
) -> None:
    config_path = path.join(vega_home, "config", "node", "config.toml")
    config_toml = toml.load(config_path)
    config_toml["Blockchain"]["Null"]["GenesisFile"] = path.join(
        vega_home, "genesis.json"
    )
    config_toml["Blockchain"]["Null"]["BlockDuration"] = block_duration
    config_toml["Blockchain"]["Null"]["TransactionsPerBlock"] = transactions_per_block

    existing_ports = set(port_config.values())
    for port in Ports:
        if port in port_config:
            continue
        new_port = find_free_port(existing_ports)
        existing_ports.add(new_port)
        port_config[port] = new_port

    with open(config_path, "w") as f:
        toml.dump(config_toml, f)

    for port_key, update_configs in PORT_UPDATERS.items():
        for config in update_configs:
            file_path = path.join(vega_home, *config.file_path)
            config_toml = toml.load(file_path)
            elem = config_toml
            for k in config.config_path:
                elem = elem[k]
            elem[config.key] = config.val_func(port_config[port_key])

            with open(file_path, "w") as f:
                toml.dump(config_toml, f)


def manage_vega_processes(
    vega_path: str,
    data_node_path: str,
    vega_wallet_path: str,
    vega_console_path: Optional[str] = None,
    run_with_console: bool = False,
    port_config: Optional[Dict[Ports, int]] = None,
    transactions_per_block: int = 1,
    block_duration: str = "1s",
    run_wallet: bool = False,
    retain_log_files: bool = False,
    log_dir: Optional[str] = None,
    replay_from_path: Optional[str] = None,
    store_transactions: bool = True,
) -> None:
    logging.basicConfig(level=logging.INFO)
    port_config = port_config if port_config is not None else {}

    # Explicitly not using context here so that crashed logs are retained
    tmp_vega_dir = tempfile.mkdtemp(prefix="vega-sim-") if log_dir is None else log_dir
    logger.info(f"Running NullChain from vegahome of {tmp_vega_dir}")
    if port_config.get(Ports.DATA_NODE_GRAPHQL):
        logger.info(
            f"Launching GraphQL node at port {port_config.get(Ports.DATA_NODE_GRAPHQL)}"
        )
    if port_config.get(Ports.CONSOLE):
        logger.info(f"Launching Console at port {port_config.get(Ports.CONSOLE)}")

    dest_dir = f"{tmp_vega_dir}/vegahome"
    shutil.copytree(vega_home_path, dest_dir)
    for dirpath, _, filenames in os.walk(dest_dir):
        os.utime(dirpath, None)
        for file in filenames:
            os.utime(os.path.join(dirpath, file), None)

    tmp_vega_home = tmp_vega_dir + "/vegahome"
    _update_node_config(
        tmp_vega_home,
        port_config=port_config,
        transactions_per_block=transactions_per_block,
        block_duration=block_duration,
    )

    dataNodeProcess = _popen_process(
        [
            data_node_path,
            "start",
            "--home=" + tmp_vega_home,
            "--chainID=CUSTOM",
        ],
        dir_root=tmp_vega_dir,
        log_name="data_node",
    )

    vegaFaucetProcess = _popen_process(
        [
            vega_path,
            "faucet",
            "run",
            "--passphrase-file=" + tmp_vega_home + "/passphrase-file",
            "--home=" + tmp_vega_home,
        ],
        dir_root=tmp_vega_dir,
        log_name="faucet",
    )

    vega_args = [
        vega_path,
        "start",
        "--nodewallet-passphrase-file=" + tmp_vega_home + "/passphrase-file",
        "--home=" + tmp_vega_home,
    ]

    if store_transactions:
        replay_file = (
            replay_from_path
            if replay_from_path is not None
            else tmp_vega_home + "/replay"
        )
        vega_args.extend(
            [
                f"--blockchain.nullchain.replay-file={replay_file}",
                "--blockchain.nullchain.record",
            ]
        )
    if replay_from_path is not None:
        vega_args.extend(
            [
                f"--blockchain.nullchain.replay-file={replay_from_path}",
                "--blockchain.nullchain.replay",
            ]
        )

    vegaNodeProcess = _popen_process(
        vega_args,
        dir_root=tmp_vega_dir,
        log_name="node",
    )

    processes = [dataNodeProcess, vegaFaucetProcess, vegaNodeProcess]

    if run_wallet:
        for _ in range(3000):
            try:
                requests.get(
                    f"http://localhost:{port_config.get(Ports.DATA_NODE_REST)}/time"
                ).raise_for_status()
                requests.get(
                    f"http://localhost:{port_config.get(Ports.CORE_REST)}/blockchain/height"
                ).raise_for_status()
                break
            except (
                MaxRetryError,
                requests.exceptions.ConnectionError,
                requests.exceptions.HTTPError,
            ):
                time.sleep(0.1)
        wallet_args = [
            vega_wallet_path,
            "service",
            "run",
            "--network",
            "local",
            "--home=" + tmp_vega_home,
            "--automatic-consent",
        ]

        vegaWalletProcess = _popen_process(
            wallet_args,
            dir_root=tmp_vega_dir,
            log_name="vegawallet",
        )
        processes.append(vegaWalletProcess)

    if run_with_console:
        env_copy = os.environ.copy()
        env_copy.update(
            {
                "NX_VEGA_URL": (
                    f"http://localhost:{port_config[Ports.DATA_NODE_GRAPHQL]}/query"
                ),
                "NX_VEGA_WALLET_URL": f"http://localhost:{port_config[Ports.WALLET]}",
                "NX_VEGA_ENV": "CUSTOM",
                "NX_PORT": f"{port_config[Ports.CONSOLE]}",
                "NODE_ENV": "development",
                "NX_VEGA_NETWORKS": "{}",
            }
        )
        console_process = _popen_process(
            [
                "yarn",
                "--cwd",
                vega_console_path,
                "nx",
                "serve",
                "--port",
                f"{port_config[Ports.CONSOLE]}",
                "-o",
                "trading",
            ],
            dir_root=tmp_vega_dir,
            log_name="console",
            env=env_copy,
        )
        processes.append(console_process)

    signal.sigwait([signal.SIGKILL, signal.SIGTERM])
    for process in processes:
        process.terminate()
    for process in processes:
        return_code = process.poll()
        if return_code is not None:
            continue
        # Could mean 5s wait per process, but we're not holding the outer process
        # and would really be a symptom of these children taking too long to close
        time.sleep(5)
        process.kill()

    if not retain_log_files:
        shutil.rmtree(tmp_vega_dir)


class VegaServiceNull(VegaService):
    PORT_TO_FIELD_MAP = {
        Ports.WALLET: "wallet_port",
        Ports.DATA_NODE_GRPC: "data_node_grpc_port",
        Ports.DATA_NODE_REST: "data_node_rest_port",
        Ports.DATA_NODE_GRAPHQL: "data_node_graphql_port",
        Ports.DATA_NODE_POSTGRES: "data_node_postgres_port",
        Ports.FAUCET: "faucet_port",
        Ports.VEGA_NODE: "vega_node_port",
        Ports.CORE_GRPC: "vega_node_grpc_port",
        Ports.CORE_REST: "vega_node_rest_port",
        Ports.CONSOLE: "console_port",
    }

    def __init__(
        self,
        vega_path: Optional[str] = None,
        data_node_path: Optional[str] = None,
        vega_wallet_path: Optional[str] = None,
        vega_console_path: Optional[str] = None,
        start_immediately: bool = False,
        run_with_console: bool = False,
        run_wallet_with_token_dapp: bool = False,
        port_config: Optional[Dict[Ports, int]] = None,
        warn_on_raw_data_access: bool = True,
        transactions_per_block: int = 1,
        seconds_per_block: int = 1,
        use_full_vega_wallet: bool = False,
        start_order_feed: bool = True,
        retain_log_files: bool = False,
        launch_graphql: bool = False,
        store_transactions: bool = True,
        replay_from_path: Optional[str] = None,
    ):
        super().__init__(
            can_control_time=True,
            warn_on_raw_data_access=warn_on_raw_data_access,
            seconds_per_block=seconds_per_block,
        )
        self.retain_log_files = retain_log_files

        self.vega_path = vega_path or path.join(vega_bin_path, "vega")
        self.data_node_path = data_node_path or path.join(vega_bin_path, "data-node")
        self.vega_wallet_path = vega_wallet_path or path.join(
            vega_bin_path, "vegawallet"
        )
        self.vega_console_path = vega_console_path or path.join(
            vega_bin_path, "console"
        )
        self.proc = None
        self.run_with_console = run_with_console
        self.run_wallet_with_token_dapp = run_wallet_with_token_dapp

        self.transactions_per_block = transactions_per_block
        self.seconds_per_block = seconds_per_block

        self._wallet = None
        self._use_full_vega_wallet = use_full_vega_wallet
        self.store_transactions = store_transactions

        self.log_dir = tempfile.mkdtemp(prefix="vega-sim-")

        self._start_order_feed = start_order_feed
        self.launch_graphql = launch_graphql
        self.replay_from_path = replay_from_path

        if port_config is None:
            self._assign_ports()
        else:
            for key, name in self.PORT_TO_FIELD_MAP.items():
                setattr(self, name, port_config[key])

        if start_immediately:
            self.start()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, type, value, traceback):
        self.stop()

    def wait_fn(self, wait_multiple: float = 1) -> None:
        self.wait_for_core_catchup()
        self.forward(f"{int(wait_multiple * self.seconds_per_block)}s")
        self.wait_for_core_catchup()

    @property
    def wallet(self) -> Wallet:
        if self._wallet is None:
            if self._use_full_vega_wallet:
                self._wallet = VegaWallet(self.wallet_url)
            else:
                self._wallet = SlimWallet(
                    self.core_client,
                    full_wallet=VegaWallet(self.wallet_url)
                    if self.run_with_console
                    else None,
                    log_dir=self.log_dir,
                )
        return self._wallet

    def _assign_ports(self):
        self.wallet_port = 0
        self.data_node_rest_port = 0
        self.data_node_grpc_port = 0
        self.data_node_graphql_port = 0
        self.data_node_postgres_port = 0
        self.faucet_port = 0
        self.vega_node_port = 0
        self.vega_node_grpc_port = 0
        self.vega_node_rest_port = 0
        self.console_port = 0
        for port_opt in self.PORT_TO_FIELD_MAP.values():
            curr_ports = set(
                [getattr(self, port) for port in self.PORT_TO_FIELD_MAP.values()]
            )
            setattr(self, port_opt, find_free_port(curr_ports))

    def _check_started(self) -> None:
        if self.proc is None:
            raise ServiceNotStartedError("NullChain Vega accessed without starting")

    def _generate_port_config(self) -> Dict[Ports, int]:
        return {
            Ports.WALLET: self.wallet_port,
            Ports.DATA_NODE_GRPC: self.data_node_grpc_port,
            Ports.DATA_NODE_REST: self.data_node_rest_port,
            Ports.DATA_NODE_GRAPHQL: self.data_node_graphql_port,
            Ports.DATA_NODE_POSTGRES: self.data_node_postgres_port,
            Ports.FAUCET: self.faucet_port,
            Ports.VEGA_NODE: self.vega_node_port,
            Ports.CORE_GRPC: self.vega_node_grpc_port,
            Ports.CORE_REST: self.vega_node_rest_port,
            Ports.CONSOLE: self.console_port,
        }

    def start(self, block_on_startup: bool = True) -> None:
        ctx = multiprocessing.get_context()
        port_config = self._generate_port_config()
        self.proc = ctx.Process(
            target=manage_vega_processes,
            kwargs={
                "vega_path": self.vega_path,
                "data_node_path": self.data_node_path,
                "vega_wallet_path": self.vega_wallet_path,
                "vega_console_path": self.vega_console_path,
                "run_with_console": self.run_with_console,
                "port_config": port_config,
                "transactions_per_block": self.transactions_per_block,
                "block_duration": f"{int(self.seconds_per_block)}s",
                "run_wallet": self._use_full_vega_wallet or self.run_with_console,
                "retain_log_files": self.retain_log_files,
                "log_dir": self.log_dir,
                "store_transactions": self.store_transactions,
                "replay_from_path": self.replay_from_path,
            },
        )
        self.proc.start()

        if self.run_with_console:
            logger.info(
                "Vega Running. Console launched at"
                f" http://localhost:{self.console_port}"
            )

        if block_on_startup:
            # Wait for startup
            started = False
            for _ in range(600):
                try:
                    requests.get(
                        f"http://localhost:{self.data_node_rest_port}/time"
                    ).raise_for_status()
                    requests.get(
                        f"http://localhost:{self.vega_node_rest_port}/blockchain/height"
                    ).raise_for_status()
                    if self.run_with_console or self._use_full_vega_wallet:
                        requests.get(
                            f"http://localhost:{self.wallet_port}/api/v1/status"
                        ).raise_for_status()

                    started = True
                    break
                except (
                    MaxRetryError,
                    requests.exceptions.ConnectionError,
                    requests.exceptions.HTTPError,
                ):
                    time.sleep(0.1)
            if not started:
                raise VegaStartupTimeoutError(
                    "Timed out waiting for Vega simulator to start up"
                )

        if self.run_with_console:
            webbrowser.open(f"http://localhost:{port_config[Ports.CONSOLE]}/", new=2)

        if self.launch_graphql:
            webbrowser.open(
                f"http://localhost:{port_config[Ports.DATA_NODE_GRAPHQL]}/", new=2
            )

        if self._start_order_feed:
            self.start_order_monitoring()

    # Class internal as at some point the host may vary as well as the port
    @staticmethod
    def _build_url(port: int, prefix: str = "http://"):
        return f"{prefix}localhost:{port}"

    def stop(self) -> None:
        super().stop()
        if self.proc is None:
            logger.info("Stop called but nothing to stop")
        else:
            self.proc.terminate()

    @property
    def wallet_url(self) -> str:
        return self._build_url(self.wallet_port)

    @property
    def data_node_rest_url(self) -> str:
        return self._build_url(self.data_node_rest_port)

    @property
    def data_node_grpc_url(self) -> str:
        return self._build_url(self.data_node_grpc_port, prefix="")

    @property
    def faucet_url(self) -> str:
        return self._build_url(self.faucet_port)

    @property
    def vega_node_url(self) -> str:
        return self._build_url(self.vega_node_port)

    @property
    def vega_node_grpc_url(self) -> str:
        return self._build_url(self.vega_node_grpc_port, prefix="")

    def clone(self) -> VegaServiceNull:
        """Creates a clone of the service without the handle to other processes.

        This is required as when spinning a Nullchain service out into
        separate processes we need to start the various components in the main
        thread (as daemon processes cannot spawn daemon processes), however want
        to maintain a handle to these in the child.
        """
        return VegaServiceNull(
            self.vega_path,
            self.data_node_path,
            self.vega_wallet_path,
            start_immediately=False,
            port_config=self._generate_port_config(),
            use_full_vega_wallet=self._use_full_vega_wallet,
            warn_on_raw_data_access=self.warn_on_raw_data_access,
        )



from __future__ import annotations
from asyncio import wait_for

import logging
import uuid
from time import time
from typing import Callable, List, Optional, Tuple, Union

import vega_sim.grpc.client as vac
import vega_sim.proto.data_node.api.v2 as data_node_protos_v2

import vega_sim.api.data as data
import vega_sim.api.data_raw as data_raw


import vega_sim.proto.vega as vega_protos
from vega_sim.proto.vega.commands.v1.commands_pb2 import (
    OrderCancellation,
    OrderAmendment,
    OrderSubmission,
)
from vega_sim.api.helpers import (
    ProposalNotAcceptedError,
    enum_to_str,
    get_enum,
    wait_for_acceptance,
    wait_for_core_catchup,
)
from vega_sim.wallet.base import Wallet

logger = logging.getLogger(__name__)


class OrderRejectedError(Exception):
    pass


def submit_order(
    wallet_name: str,
    data_client: vac.VegaTradingDataClientV2,
    wallet: Wallet,
    market_id: str,
    order_type: Union[vega_protos.vega.Order.Type, str],
    time_in_force: Union[vega_protos.vega.Order.TimeInForce, str],
    side: Union[vega_protos.vega.Side, str],
    volume: float,
    price: Optional[str] = None,
    expires_at: Optional[int] = None,
    pegged_order: Optional[vega_protos.vega.PeggedOrder] = None,
    wait: bool = True,
    time_forward_fn: Optional[Callable[[], None]] = None,
    order_ref: Optional[str] = None,
    key_name: Optional[str] = None,
) -> Optional[str]:
    """
    Submit orders as specified to required pre-existing market.
    Optionally wait for acceptance of order (raises on non-acceptance)

    Args:
        wallet_name:
            str, the wallet name performing the action
        data_client:
            VegaTradingDataClientV2, a gRPC data client to the vega data node
        wallet:
            Wallet, wallet client
        pub_key:
            str, pub key of the account placing the order
        market_id:
            str, the ID of the required market on vega
        order_type:
            vega.Order.Type or str, The type of order required (market/limit etc).
                See API documentation for full list of options
        time_in_force:
            vega.Order.TimeInForce or str, The time in force setting for the order
                (GTC, GTT, Fill Or Kill etc.)
                See API documentation for full list of options
        side:
            vega.Side or str, Side of the order (BUY or SELL)
        volume:
            float, volume of the order
        price:
            str, price of the order
        expires_at:
            int, Optional timestamp for when the order will expire, in
            nanoseconds since the epoch,
                required field only for [`Order.TimeInForce`].
                Defaults to 2 minutes
        pegged_order:
            vega.PeggedOrder, Used to specify the details for a pegged order
        wait:
            bool, whether to wait for order acceptance.
                If true, will raise an error if order is not accepted
        time_forward_fn:
            optional function, Function which takes no arguments and
            waits or manually forwards the chain when waiting for order acceptance
        order_ref:
            optional str, reference for later identification of order
        key_name:
            Optional[str], key name stored in metadata. Defaults to None.

    Returns:
        Optional[str], Order ID if wait is True, otherwise None
    """

    order_data = order_submission(
        data_client=data_client,
        market_id=market_id,
        size=volume,
        side=side,
        time_in_force=time_in_force,
        order_type=order_type,
        expires_at=expires_at,
        reference=order_ref,
        price=price,
        pegged_order=pegged_order,
    )

    # Sign the transaction with an order submission command
    # Note: Setting propagate to true will also submit to a Vega node
    wallet.submit_transaction(
        transaction=order_data,
        name=wallet_name,
        transaction_type="order_submission",
        key_name=key_name,
    )

    logger.debug(f"Submitted Order on {side} at price {price}.")

    if wait:

        def _proposal_loader(
            order_ref: str, market_id: str, data_client: vac.VegaTradingDataClientV2
        ) -> vega_protos.vega.Order:
            orders = data_raw.list_orders(
                market_id=market_id,
                reference=order_ref,
                data_client=data_client,
                live_only=False,
            )
            return orders[0]

        try:
            time_forward_fn()
            logger.debug("Waiting for proposal acceptance")
            response = wait_for_acceptance(
                order_data.reference,
                lambda r: _proposal_loader(r, market_id, data_client),
            )
        except ProposalNotAcceptedError:
            time_forward_fn()
            response = wait_for_acceptance(
                order_data.reference,
                lambda r: _proposal_loader(r, market_id, data_client),
            )
        order_status = enum_to_str(vega_protos.vega.Order.Status, response.status)

        logger.debug(
            f"Order processed, ID: {response.id}, Status: {order_status}, Version:"
            f" {response.version}"
        )
        if order_status == "STATUS_REJECTED":
            raise OrderRejectedError(
                "Rejection reason:"
                f" {enum_to_str(vega_protos.vega.OrderError, response.reason)}"
            )
        return response.id


def amend_order(
    wallet_name: str,
    wallet: Wallet,
    market_id: str,
    order_id: str,
    price: Optional[str] = None,
    expires_at: Optional[int] = None,
    pegged_offset: Optional[str] = None,
    pegged_reference: Optional[vega_protos.vega.PeggedReference] = None,
    volume_delta: float = 0,
    time_in_force: Optional[Union[vega_protos.vega.Order.TimeInForce, str]] = None,
    key_name: Optional[str] = None,
):
    """
    Amend a Limit order by orderID in the specified market

    Args:
        wallet_name:
            str, the wallet name performing the action
        wallet:
            Wallet, wallet client
        market_id:
            str, the ID of the required market on vega
        order_type:
            vega.Order.Type or str, The type of order required (market/limit etc).
                See API documentation for full list of options
        side:
            vega.Side or str, Side of the order (BUY or SELL)
        volume_delta:
            float, change in volume of the order
        price:
            str, price of the order
        time_in_force:
            vega.Order.TimeInForce or str, The time in force setting for the order
                (Only valid options for market are TIME_IN_FORCE_IOC and
                    TIME_IN_FORCE_FOK)
                See API documentation for full list of options
                Defaults to Fill or Kill
        key_name:
            Optional[str], key name stored in metadata. Defaults to None.
    """

    order_data = order_amendment(
        order_id=order_id,
        market_id=market_id,
        price=price,
        size_delta=volume_delta,
        expires_at=expires_at,
        time_in_force=time_in_force,
        pegged_offset=pegged_offset,
        pegged_reference=pegged_reference,
    )

    wallet.submit_transaction(
        transaction=order_data,
        name=wallet_name,
        transaction_type="order_amendment",
        key_name=key_name,
    )
    logger.debug(f"Submitted Order amendment for {order_id}.")


def cancel_order(
    wallet_name: str,
    wallet: Wallet,
    market_id: str,
    order_id: str,
    key_name: Optional[str] = None,
):
    """
    Cancel Order

    Args:
        wallet_name:
            str, the wallet name performing the action
        wallet:
            Wallet, wallet client
        market_id:
            str, the ID of the required market on vega
        order_id:
            str, Identifier of the order to cancel
    """

    order_data = order_cancellation(
        order_id=order_id,
        market_id=market_id,
    )

    wallet.submit_transaction(
        transaction=order_data,
        name=wallet_name,
        transaction_type="order_cancellation",
        key_name=key_name,
    )

    logger.debug(f"Cancelled order {order_id} on market {market_id}")


def submit_simple_liquidity(
    wallet_name: str,
    wallet: Wallet,
    market_id: str,
    commitment_amount: int,
    fee: float,
    reference_buy: str,
    reference_sell: str,
    delta_buy: int,
    delta_sell: int,
    is_amendment: bool = False,
    key_name: Optional[str] = None,
):
    """Submit/Amend a simple liquidity commitment (LP) with a single amount on each side.

    Args:
        wallet_name:
            str, the wallet name performing the action
        wallet:
            Wallet, wallet client
        market_id:
            str, The ID of the market to place the commitment on
        commitment_amount:
            int, The amount in asset decimals of market asset to commit to liquidity provision
        fee:
            float, The fee level at which to set the LP fee (in %, e.g. 0.01 == 1% and 1 == 100%)
        reference_buy:
            str, the reference point to use for the buy side of LP
        reference_sell:
            str, the reference point to use for the sell side of LP
        delta_buy:
            int, the offset from reference point for the buy side of LP
        delta_sell:
            int, the offset from reference point for the sell side of LP
    """
    return submit_liquidity(
        wallet_name=wallet_name,
        wallet=wallet,
        market_id=market_id,
        commitment_amount=commitment_amount,
        fee=fee,
        buy_specs=[(reference_buy, delta_buy, 1)],
        sell_specs=[(reference_sell, delta_sell, 1)],
        is_amendment=is_amendment,
        key_name=key_name,
    )


def submit_liquidity(
    wallet_name: str,
    wallet: Wallet,
    market_id: str,
    commitment_amount: int,
    fee: float,
    buy_specs: List[Tuple[str, int, int]],
    sell_specs: List[Tuple[str, int, int]],
    is_amendment: bool = False,
    key_name: Optional[str] = None,
):
    """Submit/Amend a custom liquidity profile.

    Args:
        wallet_name:
            str, the wallet name performing the action
        wallet:
            Wallet, wallet client
        market_id:
            str, The ID of the market to place the commitment on
        commitment_amount:
            int, The amount in asset decimals of market asset to commit
            to liquidity provision
        fee:
            float, The fee level at which to set the LP fee
             (in %, e.g. 0.01 == 1% and 1 == 100%)
        buy_specs:
            List[Tuple[str, int, int]], List of tuples, each containing a reference
            point in their first position, a desired offset in their second and
            a proportion in third
        sell_specs:
            List[Tuple[str, int, int]], List of tuples, each containing a reference
            point in their first position, a desired offset in their second and
            a proportion in third
        key_name:
            Optional[str], key name stored in metadata. Defaults to None.
    """

    if is_amendment:
        command = vega_protos.commands.v1.commands.LiquidityProvisionAmendment
        submission_name = "liquidity_provision_amendment"
    else:
        command = vega_protos.commands.v1.commands.LiquidityProvisionSubmission
        submission_name = "liquidity_provision_submission"

    submission = command(
        market_id=market_id,
        commitment_amount=str(commitment_amount),
        fee=str(fee),
        buys=[
            vega_protos.vega.LiquidityOrder(
                reference=spec[0],
                offset=str(spec[1]),
                proportion=spec[2],
            )
            for spec in buy_specs
        ],
        sells=[
            vega_protos.vega.LiquidityOrder(
                reference=spec[0],
                offset=str(spec[1]),
                proportion=spec[2],
            )
            for spec in sell_specs
        ],
    )
    wallet.submit_transaction(
        transaction=submission,
        name=wallet_name,
        transaction_type=submission_name,
        key_name=key_name,
    )
    logger.debug(f"Submitted liquidity on market {market_id}")


def pegged_order(
    reference: vega_protos.vega.PeggedReference,
    offset: str,
) -> vega_protos.vega.PeggedOrder:
    """Creates a Vega PeggedOrder object.

    Args:
        reference (vega_protos.vega.PeggedReference):
            Reference to offset price from.
        offset (str):
            Value to offset price from reference.

    Returns:
        vega_protos.vega.PeggedOrder:
            The created Vega PeggedOrder object.
    """
    return vega_protos.vega.PeggedOrder(
        reference=reference,
        offset=offset,
    )


def order_amendment(
    order_id: str,
    market_id: str,
    price: str,
    size_delta: int,
    expires_at: Optional[int] = None,
    time_in_force: Optional[vega_protos.vega.Order.TimeInForce] = None,
    pegged_offset: Optional[str] = None,
    pegged_reference: Optional[vega_protos.vega.PeggedReference] = None,
) -> OrderAmendment:
    """Creates a Vega OrderAmendment object.

    Args:
        order_id (str):
            Id of order to amend.
        market_id (str):
            Id of market containing order to amend.
        price (str):
            New price of the order.
        size_delta (int):
            Amount to change the order size by.
        expires_at (Optional[int]):
            Timestamp of order expiry. Defaults to None.
        time_in_force (Optional[vega_protos.vega.Order.TimeInForce]):
            New time_in_force option for order. Defaults to None.
        pegged_offset (Optional[str]):
            New amount to offset order price by from reference. Defaults to None.
        pegged_reference (Optional[vega_protos.vega.PeggedReference]):
            New reference to offset order price from. Defaults to None.

    Returns:
        OrderAmendment:
            The created Vega OrderAmendment object.
    """

    time_in_force = (
        time_in_force
        if time_in_force is not None
        else vega_protos.vega.Order.TimeInForce.TIME_IN_FORCE_UNSPECIFIED
    )

    time_in_force = get_enum(time_in_force, vega_protos.vega.Order.TimeInForce)

    if pegged_reference is not None:
        pegged_reference = get_enum(pegged_reference, vega_protos.vega.PeggedReference)

    command = OrderAmendment(
        order_id=order_id,
        market_id=market_id,
        size_delta=size_delta,
        time_in_force=time_in_force,
    )

    # Update OrderSubmission object with optional fields if specified
    for attr, val in [
        ("price", price),
        ("expires_at", expires_at),
        ("pegged_offset", pegged_offset),
        ("pegged_reference", pegged_reference),
    ]:
        if val is not None:
            setattr(command, attr, val)

    # Return the created and updated OrderSubmission object
    return command


def order_cancellation(
    order_id: str,
    market_id: str,
) -> OrderCancellation:
    """Creates a Vega OrderCancellation object.

    Args:
        order_id (str):
            Id of order to cancel.
        market_id (str):
            Id of market containing order to cancel.

    Returns:
        OrderCancellation:
            The created Vega OrderCancellation object.
    """
    return OrderCancellation(
        order_id=order_id,
        market_id=market_id,
    )


def order_submission(
    data_client: vac.VegaTradingDataClient,
    market_id: str,
    size: int,
    side: Union[vega_protos.vega.Side, str],
    time_in_force: Union[vega_protos.vega.Order.TimeInForce, str],
    order_type: Union[vega_protos.vega.Order.Type, str],
    expires_at: Optional[int] = None,
    reference: Optional[str] = None,
    price: Optional[str] = None,
    pegged_order: Optional[Union[vega_protos.PeggedOrder, str]] = None,
) -> OrderSubmission:
    """Creates a Vega OrderSubmission object.

    Args:
        data_client (vac.VegaTradingDataClient):
            Client for trading data api.
        market_id (str):
            Id of market to place order in.
        size (int):
            Size of order to be placed.
        side (Union[vega_protos.vega.Side, str]):
            Side of order to be placed, "SIDE_BUY" or "SIDE_SELL".
        time_in_force (Union[vega_protos.vega.Order.TimeInForce, str]):
            Time in force option for order.
        order_type (Union[vega_protos.vega.Order.Type, str]):
            Type of order, "TYPE_LIMIT" or "TYPE_MARKET".
        expires_at (Optional[int]):
            Determines timestamp at which order expires, only required for orders of
            "TYPE_LIMIT" and "TIME_IN_FORCE_GTT". Defaults to None.
        reference (Optional[str]):
            Reference id to use for order. Defaults to None.
        price (Optional[str]):
            Price to place order at, only required for "TYPE_LIMIT". Defaults to None.
        pegged_order (Optional[Union[vega_protos.PeggedOrder, str]]):
            PeggedOrder object defining whether order is pegged. Defaults to None.

    Returns:
        OrderSubmission:
            The created Vega OrderSubmission object.
    """

    side = get_enum(side, vega_protos.vega.Side)
    order_type = get_enum(order_type, vega_protos.vega.Order.Type)
    time_in_force = get_enum(time_in_force, vega_protos.vega.Order.TimeInForce)

    # Ensure no expires_at field set if TIF is not TIME_IN_FORCE_GTT
    if time_in_force != vega_protos.vega.Order.TimeInForce.TIME_IN_FORCE_GTT:
        expires_at = None
    # Ensure an expires_at field set if TIF is TIME_IN_FORCE_GTT
    elif expires_at is None:
        blockchain_time = data_client.GetVegaTime(
            data_node_protos_v2.trading_data.GetVegaTimeRequest()
        ).timestamp
        expires_at = int(blockchain_time + 120 * 1e9)  # expire in 2 minutes

    reference = str(uuid.uuid4()) if reference is None else reference

    # Create OrderSubmission object with required fields
    command = OrderSubmission(
        market_id=market_id,
        size=size,
        side=side,
        time_in_force=time_in_force,
        type=order_type,
    )

    # Update OrderSubmission object with optional fields if specified
    for attr, val in [
        ("price", price),
        ("expires_at", expires_at),
        ("reference", reference),
    ]:
        if val is not None:
            setattr(command, attr, val)

    if pegged_order is not None:
        command.pegged_order.CopyFrom(pegged_order)

    # Return the created and updated OrderSubmission object
    return command


def batch_market_instructions(
    wallet: Wallet,
    wallet_name: str,
    key_name: Optional[str] = None,
    amendments: Optional[List[OrderAmendment]] = [],
    submissions: Optional[List[OrderSubmission]] = [],
    cancellations: Optional[List[OrderCancellation]] = [],
):
    """Submits a batch of market instructions.

    Args:
        wallet (Wallet):
            Wallet client used to submit transaction.
        wallet_name (str):
            Name of wallet to submit transaction.
        key_name (Optional[str], optional):
            Name of key to submit transaction. Defaults to None.
        amendments (Optional[List[OrderAmendment]]):
            List of OrderAmendment objects to process sequentially. Defaults to [].
        submissions (Optional[List[OrderSubmission]]):
            List of OrderSubmission objects to process sequentially. Defaults to [].
        cancellations (Optional[ List[OrderCancellation] ]):
            List of OrderCancellation objects to process sequentially. Defaults to [].
    """

    command = vega_protos.commands.v1.commands.BatchMarketInstructions(
        submissions=submissions, amendments=amendments, cancellations=cancellations
    )

    wallet.submit_transaction(
        transaction=command,
        name=wallet_name,
        transaction_type="batch_market_instructions",
        key_name=key_name,
    )
    logger.debug(
        f"Submitted a batch of {len(cancellations)} cancellation, {len(amendments)} amendment, and {len(submissions)} submission instructions."
    )



import json
import logging
from typing import Callable, Optional

import vega_sim.api.data_raw as data_raw
import vega_sim.grpc.client as vac
import vega_sim.proto.data_node.api.v2 as data_node_protos_v2

import vega_sim.proto.vega as vega_protos
import vega_sim.proto.vega.commands.v1 as commands_protos
import vega_sim.proto.vega.data.v1 as oracles_protos
import vega_sim.proto.vega.data_source_pb2 as data_source_protos
from vega_sim.api.data import find_asset_id
from vega_sim.api.helpers import (
    ProposalNotAcceptedError,
    enum_to_str,
    generate_id,
    wait_for_acceptance,
)
from vega_sim.api.market import MarketConfig
from vega_sim.proto.vega.commands.v1.commands_pb2 import ProposalSubmission
from vega_sim.wallet.base import Wallet

logger = logging.getLogger(__name__)

ASSET_URL_BASE = "{node_url}/assets"


class LowBalanceError(Exception):
    pass


def _proposal_loader(
    proposal_ref: str,
    data_client: vac.VegaTradingDataClientV2,
) -> bool:
    request = data_node_protos_v2.trading_data.GetGovernanceDataRequest(
        reference=proposal_ref,
    )
    response = data_client.GetGovernanceData(request)
    return response.data


def _default_risk_model() -> vega_protos.markets.LogNormalRiskModel:
    return vega_protos.markets.LogNormalRiskModel(
        risk_aversion_parameter=0.01,
        tau=1.90128526884173e-06,
        params=vega_protos.markets.LogNormalModelParams(mu=0, r=0.016, sigma=3.0),
    )


def _default_price_monitoring_parameters() -> (
    vega_protos.markets.PriceMonitoringParameters
):
    return vega_protos.markets.PriceMonitoringParameters(
        triggers=[
            vega_protos.markets.PriceMonitoringTrigger(
                # in seconds, so 24h, the longer the wider bounds
                horizon=24 * 3600,
                # number close to but below 1 leads to wide bounds
                probability="0.999999",
                # in seconds
                auction_extension=5,
            )
        ]
    )


def get_blockchain_time(data_client: vac.VegaTradingDataClientV2) -> int:
    """Returns blockchain time in seconds since the epoch"""
    blockchain_time = data_client.GetVegaTime(
        data_node_protos_v2.trading_data.GetVegaTimeRequest()
    ).timestamp
    return int(blockchain_time / 1e9)


def propose_market_from_config(
    data_client: vac.VegaTradingDataClientV2,
    wallet: Wallet,
    proposal_wallet_name: str,
    market_config: MarketConfig,
    closing_time: str,
    enactment_time: str,
    time_forward_fn: Optional[Callable[[], None]] = None,
    governance_asset: Optional[str] = "VOTE",
    proposal_key_name: Optional[str] = None,
) -> str:
    # Make sure Vega network has governance asset
    vote_asset_id = find_asset_id(
        governance_asset, raise_on_missing=True, data_client=data_client
    )
    pub_key = wallet.public_key(proposal_wallet_name, proposal_key_name)

    # Request accounts for party and check governance asset balance
    party_accounts = data_raw.party_accounts(
        data_client=data_client, asset_id=vote_asset_id, party_id=pub_key
    )

    voting_balance = 0
    for account in party_accounts:
        if account.asset == vote_asset_id:
            voting_balance = account.balance
            break

    if voting_balance == 0:
        raise LowBalanceError(
            f"Public key {pub_key} is missing governance token {governance_asset}"
        )

    # Build NewMarketConfiguration proto
    changes = market_config.build()

    # Build ProposalTerms proto
    proposal = _build_generic_proposal(
        pub_key=pub_key,
        data_client=data_client,
        closing_time=closing_time,
        enactment_time=enactment_time,
    )
    proposal.terms.new_market.CopyFrom(changes)

    return _make_and_wait_for_proposal(
        wallet=wallet,
        wallet_name=proposal_wallet_name,
        key_name=proposal_key_name,
        proposal=proposal,
        data_client=data_client,
        time_forward_fn=time_forward_fn,
    ).proposal.id


def propose_future_market(
    market_name: str,
    wallet_name: str,
    wallet: Wallet,
    settlement_asset_id: str,
    data_client: vac.VegaTradingDataClientV2,
    termination_pub_key: str,
    governance_asset: str = "VOTE",
    future_asset: str = "BTC",
    position_decimals: Optional[int] = None,
    market_decimals: Optional[int] = None,
    closing_time: Optional[int] = None,
    enactment_time: Optional[int] = None,
    risk_model: Optional[vega_protos.markets.LogNormalRiskModel] = None,
    time_forward_fn: Optional[Callable[[], None]] = None,
    price_monitoring_parameters: Optional[
        vega_protos.markets.PriceMonitoringParameters
    ] = None,
    key_name: Optional[str] = None,
) -> str:
    """Propose a future market as specified user.

    Args:
        market_name:
            str, name of the market
        wallet_name:
            str, the wallet name performing the action
        wallet:
            Wallet, wallet client
        login_token:
            str, the token returned from proposer wallet login
        settlement_asset_id:
            str, the asset id the market will use for settlement
        data_client:
            VegaTradingDataClientV2, an instantiated gRPC client for interacting with the
                Vega data node
        termination_pub_key:
            str, the public key of the oracle to be used for trading termination
        governance_asset:
            str, the governance asset on the market
        future_asset:
            str, the symbol of the future asset used
                (used for generating/linking names of oracles)
        position_decimals:
            int, the decimal place precision to use for positions
            (e.g. 2 means 2dp, so 200 => 2.00, 3 would mean 200 => 0.2)
        market_decimals:
            int, the decimal place precision to use for market prices
            (e.g. 2 means 2dp, so 200 => 2.00, 3 would mean 200 => 0.2)
        risk_model:
            LogNormalRiskModel, A parametrised risk model on which the market will run
        price_monitoring_parameters:
            PriceMonitoringParameters, A set of parameters determining when the market
                will drop into a price auction. If not passed defaults to a very
                permissive setup
        key_name:
            Optional[str], key name stored in metadata. Defaults to None.

    Returns:
        str, the ID of the future market proposal on chain
    """
    # Make sure Vega network has governance asset
    vote_asset_id = find_asset_id(
        governance_asset, raise_on_missing=True, data_client=data_client
    )
    pub_key = wallet.public_key(wallet_name, key_name)

    # Request accounts for party and check governance asset balance
    party_accounts = data_raw.party_accounts(
        data_client=data_client, asset_id=vote_asset_id, party_id=pub_key
    )

    voting_balance = 0
    for account in party_accounts:
        if account.asset == vote_asset_id:
            voting_balance = account.balance
            break

    if voting_balance == 0:
        raise LowBalanceError(
            f"Public key {pub_key} is missing governance token {governance_asset}"
        )

    risk_model = risk_model if risk_model is not None else _default_risk_model()
    price_monitoring_parameters = (
        price_monitoring_parameters
        if price_monitoring_parameters is not None
        else _default_price_monitoring_parameters()
    )

    data_source_spec_for_settlement_data = data_source_protos.DataSourceDefinition(
        external=data_source_protos.DataSourceDefinitionExternal(
            oracle=data_source_protos.DataSourceSpecConfiguration(
                signers=[
                    oracles_protos.data.Signer(
                        pub_key=oracles_protos.data.PubKey(key=termination_pub_key)
                    )
                ],
                filters=[
                    oracles_protos.spec.Filter(
                        key=oracles_protos.spec.PropertyKey(
                            name=f"price.{future_asset}.value",
                            type=oracles_protos.spec.PropertyKey.Type.TYPE_INTEGER,
                        ),
                        conditions=[],
                    )
                ],
            )
        )
    )
    data_source_spec_for_trading_termination = data_source_protos.DataSourceDefinition(
        external=data_source_protos.DataSourceDefinitionExternal(
            oracle=data_source_protos.DataSourceSpecConfiguration(
                signers=[
                    oracles_protos.data.Signer(
                        pub_key=oracles_protos.data.PubKey(key=termination_pub_key)
                    )
                ],
                filters=[
                    oracles_protos.spec.Filter(
                        key=oracles_protos.spec.PropertyKey(
                            name="trading.terminated",
                            type=oracles_protos.spec.PropertyKey.Type.TYPE_BOOLEAN,
                        ),
                        conditions=[],
                    )
                ],
            )
        )
    )

    price_decimals = 5 if market_decimals is None else market_decimals
    market_proposal = vega_protos.governance.NewMarket(
        changes=vega_protos.governance.NewMarketConfiguration(
            instrument=vega_protos.governance.InstrumentConfiguration(
                name=market_name,
                code=market_name,
                future=vega_protos.governance.FutureProduct(
                    settlement_asset=settlement_asset_id,
                    quote_name=future_asset,
                    data_source_spec_for_settlement_data=data_source_spec_for_settlement_data,
                    data_source_spec_for_trading_termination=data_source_spec_for_trading_termination,
                    data_source_spec_binding=vega_protos.markets.DataSourceSpecToFutureBinding(
                        settlement_data_property=f"price.{future_asset}.value",
                        trading_termination_property="trading.terminated",
                    ),
                    settlement_data_decimals=price_decimals,
                ),
            ),
            decimal_places=price_decimals,
            position_decimal_places=0
            if position_decimals is None
            else position_decimals,
            metadata=[
                f"base:{future_asset}",
            ],
            liquidity_monitoring_parameters=vega_protos.markets.LiquidityMonitoringParameters(
                target_stake_parameters=vega_protos.markets.TargetStakeParameters(
                    time_window=3600, scaling_factor=1
                ),
                triggering_ratio=0.7,
                auction_extension=0,
            ),
            price_monitoring_parameters=price_monitoring_parameters,
            log_normal=risk_model,
        ),
    )

    proposal = _build_generic_proposal(
        pub_key=pub_key,
        data_client=data_client,
        closing_time=closing_time,
        enactment_time=enactment_time,
    )
    proposal.terms.new_market.CopyFrom(market_proposal)

    return _make_and_wait_for_proposal(
        wallet_name=wallet_name,
        wallet=wallet,
        proposal=proposal,
        data_client=data_client,
        time_forward_fn=time_forward_fn,
        key_name=key_name,
    ).proposal.id


def propose_network_parameter_change(
    parameter: str,
    value: str,
    wallet_name: str,
    wallet: Wallet,
    closing_time: Optional[int] = None,
    enactment_time: Optional[int] = None,
    data_client: Optional[vac.VegaTradingDataClientV2] = None,
    time_forward_fn: Optional[Callable[[], None]] = None,
    key_name: Optional[str] = None,
):
    network_param_update = _build_generic_proposal(
        pub_key=wallet.public_key(wallet_name, key_name),
        data_client=data_client,
        closing_time=closing_time,
        enactment_time=enactment_time,
    )
    network_param_update.terms.update_network_parameter.CopyFrom(
        vega_protos.governance.UpdateNetworkParameter(
            changes=vega_protos.vega.NetworkParameter(key=parameter, value=value)
        )
    )
    return _make_and_wait_for_proposal(
        wallet_name=wallet_name,
        wallet=wallet,
        proposal=network_param_update,
        data_client=data_client,
        time_forward_fn=time_forward_fn,
        key_name=key_name,
    ).proposal.id


def propose_market_update(
    market_id: str,
    wallet_name: str,
    wallet: Wallet,
    market_update: vega_protos.governance.UpdateMarketConfiguration,
    closing_time: Optional[int] = None,
    enactment_time: Optional[int] = None,
    data_client: Optional[vac.VegaTradingDataClientV2] = None,
    time_forward_fn: Optional[Callable[[], None]] = None,
    key_name: Optional[str] = None,
) -> str:
    network_param_update = _build_generic_proposal(
        pub_key=wallet.public_key(wallet_name, key_name),
        data_client=data_client,
        closing_time=closing_time,
        enactment_time=enactment_time,
    )
    network_param_update.terms.update_market.CopyFrom(
        vega_protos.governance.UpdateMarket(market_id=market_id, changes=market_update)
    )

    return _make_and_wait_for_proposal(
        wallet_name=wallet_name,
        wallet=wallet,
        proposal=network_param_update,
        data_client=data_client,
        time_forward_fn=time_forward_fn,
    ).proposal.id


def approve_proposal(
    wallet_name: str,
    proposal_id: str,
    wallet: Wallet,
    key_name: Optional[str] = None,
):
    wallet.submit_transaction(
        transaction=commands_protos.commands.VoteSubmission(
            value=vega_protos.governance.Vote.Value.VALUE_YES, proposal_id=proposal_id
        ),
        name=wallet_name,
        transaction_type="vote_submission",
        key_name=key_name,
    )


def propose_asset(
    wallet_name: str,
    wallet: Wallet,
    name: str,
    symbol: str,
    decimals: int,
    data_client: vac.VegaTradingDataClientV2,
    quantum: int = 1,
    max_faucet_amount: int = 10e9,
    closing_time: Optional[int] = None,
    enactment_time: Optional[int] = None,
    validation_time: Optional[int] = None,
    time_forward_fn: Optional[Callable[[], None]] = None,
    key_name: Optional[str] = None,
) -> str:
    asset_detail = vega_protos.assets.AssetDetails(
        name=name,
        symbol=symbol,
        decimals=decimals,
        quantum=str(int(quantum)),
        builtin_asset=vega_protos.assets.BuiltinAsset(
            max_faucet_amount_mint=str(int(max_faucet_amount))
        ),
    )
    proposal = _build_generic_proposal(
        pub_key=wallet.public_key(wallet_name, key_name),
        data_client=data_client,
        closing_time=closing_time,
        enactment_time=enactment_time,
    )
    proposal.terms.validation_timestamp = (
        validation_time
        if validation_time is not None
        else get_blockchain_time(data_client) + 10
    )

    proposal.terms.new_asset.CopyFrom(
        vega_protos.governance.NewAsset(changes=asset_detail)
    )
    return _make_and_wait_for_proposal(
        wallet_name=wallet_name,
        wallet=wallet,
        proposal=proposal,
        data_client=data_client,
        time_forward_fn=time_forward_fn,
        key_name=key_name,
    ).proposal.id


def _build_generic_proposal(
    pub_key: str,
    data_client: vac.VegaTradingDataClientV2,
    closing_time: Optional[int] = None,
    enactment_time: Optional[int] = None,
) -> commands_protos.commands.ProposalSubmission:
    # Set closing/enactment timestamps to valid time offsets
    # from the current Vega blockchain time if not already set
    none_times = [i is None for i in [closing_time, enactment_time]]
    if any(none_times):
        if not all(none_times):
            logger.warn(
                "Some times for proposal were not set. Defaulting all of them and"
                " ignoring values for those which were"
            )

        blockchain_time_seconds = get_blockchain_time(data_client)

        closing_time = blockchain_time_seconds + 172800
        enactment_time = blockchain_time_seconds + 172900

    # Propose market
    proposal_ref = f"{pub_key}-{generate_id(6)}"

    return commands_protos.commands.ProposalSubmission(
        reference=proposal_ref,
        terms=vega_protos.governance.ProposalTerms(
            closing_timestamp=closing_time,
            enactment_timestamp=enactment_time,
        ),
        rationale=vega_protos.governance.ProposalRationale(
            description="Making a proposal", title="This is a proposal"
        ),
    )


def _make_and_wait_for_proposal(
    wallet_name: str,
    wallet: Wallet,
    proposal: commands_protos.commands.ProposalSubmission,
    data_client: vac.VegaTradingDataClientV2,
    time_forward_fn: Optional[Callable[[], None]] = None,
    key_name: Optional[str] = None,
) -> ProposalSubmission:
    wallet.submit_transaction(
        transaction=proposal,
        name=wallet_name,
        transaction_type="proposal_submission",
        key_name=key_name,
    )
    logger.debug("Waiting for proposal acceptance")

    # Allow one failure, forward once more
    try:
        time_forward_fn()
        proposal = wait_for_acceptance(
            proposal.reference,
            lambda p: _proposal_loader(p, data_client),
        )
    except ProposalNotAcceptedError:
        time_forward_fn()
        proposal = wait_for_acceptance(
            proposal.reference,
            lambda p: _proposal_loader(p, data_client),
        )

    prop_state = enum_to_str(
        vega_protos.governance.Proposal.State, proposal.proposal.state
    )
    if prop_state in ["STATE_REJECTED", "STATE_DECLINED", "STATE_FAILED"]:
        raise ProposalNotAcceptedError(
            f"Your proposal was {prop_state} due to"
            f" {enum_to_str(vega_protos.governance.ProposalError, proposal.proposal.reason)}."
            f" Any further info: {proposal.proposal.error_details}"
        )
    return proposal


def settle_oracle(
    wallet_name: str,
    wallet: Wallet,
    settlement_price: float,
    oracle_name: str,
    key_name: Optional[str] = None,
) -> None:
    """
    Settle the market and send settlement price.

    Args:
        login_token:
            str, the login token for the wallet authorised to send
             termination/settlement oracle signals
        pub_key:
            str, the public key for the wallet authorised to send
             termination/settlement oracle signals
        settlement_price:
            float, final settlement price for the asset
        oracle_name:
            str, the name of the oracle to settle
        key_name:
            Optional[str], key name stored in metadata. Defaults to None.

    """

    # Use oracle feed to terminate market
    payload = {"trading.terminated": "true"}
    payload = json.dumps(payload).encode()

    oracle_submission = commands_protos.data.OracleDataSubmission(
        payload=payload,
        source=commands_protos.data.OracleDataSubmission.OracleSource.ORACLE_SOURCE_JSON,
    )

    wallet.submit_transaction(
        transaction=oracle_submission,
        name=wallet_name,
        transaction_type="oracle_data_submission",
        key_name=key_name,
    )

    # use oracle to settle market
    payload = {oracle_name: str(settlement_price)}
    payload = json.dumps(payload).encode()

    oracle_submission = commands_protos.data.OracleDataSubmission(
        payload=payload,
        source=commands_protos.data.OracleDataSubmission.OracleSource.ORACLE_SOURCE_JSON,
    )

    wallet.submit_transaction(
        transaction=oracle_submission,
        name=wallet_name,
        transaction_type="oracle_data_submission",
        key_name=key_name,
    )



"""
This is the Vega gRPC API client.

Code generated by `generate_init.py`. DO NOT EDIT.
"""

from . import (
    api,
    checkpoint,
    commands,
    events,
    data,
    snapshot,
    wallet,
    assets_pb2 as assets,
    assets_pb2_grpc as assets_grpc,
    chain_events_pb2 as chain_events,
    chain_events_pb2_grpc as chain_events_grpc,
    data_source_pb2 as data_source,
    data_source_pb2_grpc as data_source_grpc,
    governance_pb2 as governance,
    governance_pb2_grpc as governance_grpc,
    markets_pb2 as markets,
    markets_pb2_grpc as markets_grpc,
    oracle_pb2 as oracle,
    oracle_pb2_grpc as oracle_grpc,
    vega_pb2 as vega,
    vega_pb2_grpc as vega_grpc,
)

__all__ = [
    "api",
    "checkpoint",
    "commands",
    "events",
    "data",
    "snapshot",
    "tm",
    "wallet",
    "assets",
    "assets_grpc",
    "chain_events",
    "chain_events_grpc",
    "data_source",
    "data_source_grpc",
    "governance",
    "governance_grpc",
    "markets",
    "markets_grpc",
    "oracle",
    "oracle_grpc",
    "vega",
    "vega_grpc",
]



from typing import Optional
import numpy as np

from typing import Any, Dict, List, Optional, Union
import requests
import pandas as pd
from enum import Enum
import datetime

COINBASE_REQUEST_BASE = "https://api.exchange.coinbase.com/products"
COINBASE_CANDLE_BASE = COINBASE_REQUEST_BASE + "/{product_id}/candles"


class Granularity(Enum):
    MINUTE = 60
    FIVE_MINUTE = 300
    FIFTEEN_MINUTE = 900
    HOUR = 3600
    SIX_HOUR = 21600
    DAY = 86400


# Represents positions in the returned array
class CoinbaseCandle(Enum):
    TIME = 0
    LOW = 1
    HIGH = 2
    OPEN = 3
    CLOSE = 4
    VOLUME = 5


def random_walk(
    num_steps: int = 100,
    random_state: Optional[np.random.RandomState] = None,
    sigma: float = 1,
    drift: float = 0,
    starting_price: float = 100,
    decimal_precision: Optional[int] = None,
    trim_to_min: Optional[float] = None,
):
    random_state_set = random_state is not None
    random_state = random_state if random_state_set else np.random.RandomState()

    S = np.zeros(num_steps + 1)
    S[0] = starting_price

    for _ in range(100):
        dW = random_state.randn(num_steps)
        # Simulate external midprice
        for i in range(1, num_steps):
            S[i] = S[i - 1] + drift + sigma * dW[i]

        # market decimal place
        if decimal_precision:
            S = np.round(S, decimal_precision)

        # If random state is passed then error if it generates a negative price
        # Otherwise retry with a new seed

        if trim_to_min is not None:
            S[S < trim_to_min] = trim_to_min

        if (S > 0).all():
            break
        else:
            if random_state_set:
                raise Exception(
                    "Negative price generated with current random seed. Please try"
                    " another or don't specify one"
                )
            random_state = np.random.RandomState()
    if (S < 0).any():
        raise Exception("No valid price series generated after 100 attempts")
    return S


def get_trading_pairs() -> List[Dict[str, Any]]:
    headers = {"Accept": "application/json"}
    return requests.get(COINBASE_REQUEST_BASE, headers=headers).json()


def get_historic_candles(
    product_id: str,
    granularity: Optional[Granularity] = None,
    start: Optional[str] = None,
    end: Optional[str] = None,
) -> List[List[Union[int, float]]]:
    headers = {"Accept": "application/json"}
    params = {}
    if granularity is not None:
        params["granularity"] = granularity.value
    if (start is not None) or (end is not None):
        if end is None or start is None:
            raise Exception(
                "Both start and end must be specified or the specified one is ignored"
            )
        params["start"] = start
        params["end"] = end

    if start is None and end is None:
        response = requests.get(
            COINBASE_CANDLE_BASE.format(product_id=product_id),
            headers=headers,
            params=params,
        )
        response.raise_for_status()
        return response.json()

    endtime = datetime.datetime.strptime(end, "%Y-%m-%d %H:%M:%S")
    starttime = datetime.datetime.strptime(start, "%Y-%m-%d %H:%M:%S")
    res = []

    while (endtime - starttime).total_seconds() / granularity.value > 300:
        inter = starttime + datetime.timedelta(seconds=300 * granularity.value)
        params["start"] = starttime.strftime("%Y-%m-%d %H:%M:%S")
        params["end"] = inter.strftime("%Y-%m-%d %H:%M:%S")
        response = requests.get(
            COINBASE_CANDLE_BASE.format(product_id=product_id),
            headers=headers,
            params=params,
        )
        response.raise_for_status()
        res += response.json()
        starttime = inter

    params["start"] = starttime.strftime("%Y-%m-%d %H:%M:%S")
    params["end"] = endtime.strftime("%Y-%m-%d %H:%M:%S")
    response = requests.get(
        COINBASE_CANDLE_BASE.format(product_id=product_id),
        headers=headers,
        params=params,
    )
    response.raise_for_status()
    res += response.json()
    return res


def get_historic_price_series(
    product_id: str,
    price_component: CoinbaseCandle = CoinbaseCandle.CLOSE,
    granularity: Optional[Granularity] = None,
    start: Optional[str] = None,
    end: Optional[str] = None,
) -> pd.Series:
    ohlcv = get_historic_candles(
        product_id=product_id,
        granularity=granularity,
        start=start,
        end=end,
    )
    return pd.Series(
        data=[o[price_component.value] for o in ohlcv],
        index=pd.DatetimeIndex(
            [pd.to_datetime(o[CoinbaseCandle.TIME.value], unit="s") for o in ohlcv],
        ),
    ).sort_index()


if __name__ == "__main__":
    print(
        get_historic_price_series(
            "ETH-USD",
            granularity=Granularity.HOUR,
            start="2022-08-02 01:01:50",
            end="2022-09-05 09:05:20",
        )
    )


class LivePrice:
    """Iterator for getting a live product price process.

    Class is to be used when running the scenario on fairground incentives. The
    iterator can be passed to the market-maker agent and the price-sensitive
    agents to give them information regarding the live product price.

    """

    def __init__(self, product: str = "ADAUSDT"):
        self.product = product

    def __iter__(self):
        return self

    def __getitem__(self, index):
        return self._get_price()

    def __next__(self):
        return self._get_price()

    def _get_price(self):
        url = f"https://api.binance.com/api/v3/avgPrice?symbol={self.product}"
        return float(requests.get(url).json()["price"])
